r"""
Differential Form Modules

The set `\Omega^p(U, \Phi)` of `p`-forms along a differentiable manifold `U`
with values on a differentiable manifold `M` via a differentiable map
`\Phi:\ U \rightarrow M` (possibly `U = M` and `\Phi = \mathrm{Id}_M`)
is a module over the algebra `C^k(U)` of differentiable scalar fields on `U`.
It is a free module if and only if `M` is parallelizable. Accordingly,
two classes implement `\Omega^p(U, \Phi)`:

- :class:`DiffFormModule` for differential forms with values on a generic
  (in practice, not parallelizable) differentiable manifold `M`
- :class:`DiffFormFreeModule` for differential forms with values on a
  parallelizable manifold `M`
  (the subclass :class:`VectorFieldDualFreeModule` implements the special
  case of differential 1-forms on a parallelizable manifold `M`)

AUTHORS:

- Eric Gourgoulhon (2015): initial version
- Travis Scrimshaw (2016): review tweaks
- Matthias Koeppe (2022): :class:`VectorFieldDualFreeModule`

REFERENCES:

- [KN1963]_
- [Lee2013]_
"""
# *****************************************************************************
#       Copyright (C) 2015-2021 Eric Gourgoulhon <eric.gourgoulhon@obspm.fr>
#                     2016      Travis Scrimshaw <tscrimsh@umn.edu>
#                     2020      Michael Jung
#                     2022      Matthias Koeppe
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from sage.categories.modules import Modules
from sage.manifolds.differentiable.diff_form import DiffForm, DiffFormParal
from sage.manifolds.differentiable.tensorfield import TensorField
from sage.manifolds.differentiable.tensorfield_paral import TensorFieldParal
from sage.misc.cachefunc import cached_method
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.tensor.modules.ext_pow_free_module import ExtPowerDualFreeModule
from sage.tensor.modules.reflexive_module import ReflexiveModule_abstract


class DiffFormModule(UniqueRepresentation, Parent):
    r"""
    Module of differential forms of a given degree `p` (`p`-forms) along a
    differentiable manifold `U` with values on a differentiable manifold `M`.

    Given a differentiable manifold `U` and a differentiable map
    `\Phi: U \rightarrow M` to a differentiable manifold `M`, the set
    `\Omega^p(U, \Phi)` of `p`-forms along `U` with values on `M` is
    a module over `C^k(U)`, the commutative algebra of differentiable
    scalar fields on `U` (see
    :class:`~sage.manifolds.differentiable.scalarfield_algebra.DiffScalarFieldAlgebra`).
    The standard case of `p`-forms *on* a differentiable manifold `M`
    corresponds to `U = M` and `\Phi = \mathrm{Id}_M`. Other common cases
    are `\Phi` being an immersion and `\Phi` being a curve in `M`
    (`U` is then an open interval of `\RR`).

    .. NOTE::

        This class implements `\Omega^p(U,\Phi)` in the case where `M` is
        not assumed to be parallelizable; the module `\Omega^p(U, \Phi)`
        is then not necessarily free. If `M` is parallelizable, the class
        :class:`DiffFormFreeModule` must be used instead.

    INPUT:

    - ``vector_field_module`` -- module `\mathfrak{X}(U, \Phi)` of vector
      fields along `U` with values on `M` via the map `\Phi: U \rightarrow M`
    - ``degree`` -- positive integer; the degree `p` of the differential forms

    EXAMPLES:

    Module of 2-forms on a non-parallelizable 2-dimensional manifold::

        sage: M = Manifold(2, 'M')
        sage: U = M.open_subset('U') ; V = M.open_subset('V')
        sage: M.declare_union(U,V)   # M is the union of U and V
        sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
        sage: transf = c_xy.transition_map(c_uv, (x+y, x-y),
        ....:  intersection_name='W', restrictions1= x>0, restrictions2= u+v>0)
        sage: inv = transf.inverse()
        sage: W = U.intersection(V)
        sage: eU = c_xy.frame() ; eV = c_uv.frame()
        sage: XM = M.vector_field_module() ; XM
        Module X(M) of vector fields on the 2-dimensional differentiable
         manifold M
        sage: A = M.diff_form_module(2) ; A
        Module Omega^2(M) of 2-forms on the 2-dimensional differentiable
         manifold M
        sage: latex(A)
        \Omega^{2}\left(M\right)

    ``A`` is nothing but the second exterior power of the dual of ``XM``, i.e.
    we have `\Omega^{2}(M) = \Lambda^2(\mathfrak{X}(M)^*)`::

        sage: A is XM.dual_exterior_power(2)
        True

    Modules of differential forms are unique::

        sage: A is M.diff_form_module(2)
        True

    `\Omega^2(M)` is a module over the algebra `C^k(M)` of (differentiable)
    scalar fields on `M`::

        sage: A.category()
        Category of modules over Algebra of differentiable scalar fields on
         the 2-dimensional differentiable manifold M
        sage: CM = M.scalar_field_algebra() ; CM
        Algebra of differentiable scalar fields on the 2-dimensional
         differentiable manifold M
        sage: A in Modules(CM)
        True
        sage: A.base_ring() is CM
        True
        sage: A.base_module()
        Module X(M) of vector fields on the 2-dimensional differentiable
         manifold M
        sage: A.base_module() is XM
        True

    Elements can be constructed from ``A()``. In particular, ``0`` yields
    the zero element of ``A``::

        sage: z = A(0) ; z
        2-form zero on the 2-dimensional differentiable manifold M
        sage: z.display(eU)
        zero = 0
        sage: z.display(eV)
        zero = 0
        sage: z is A.zero()
        True

    while nonzero elements are constructed by providing their components in a
    given vector frame::

        sage: a = A([[0,3*x],[-3*x,0]], frame=eU, name='a') ; a
        2-form a on the 2-dimensional differentiable manifold M
        sage: a.add_comp_by_continuation(eV, W, c_uv) # finishes initializ. of a
        sage: a.display(eU)
        a = 3*x dx∧dy
        sage: a.display(eV)
        a = (-3/4*u - 3/4*v) du∧dv

    An alternative is to construct the 2-form from an empty list of
    components and to set the nonzero nonredundant components afterwards::

        sage: a = A([], name='a')
        sage: a[eU,0,1] = 3*x
        sage: a.add_comp_by_continuation(eV, W, c_uv)
        sage: a.display(eU)
        a = 3*x dx∧dy
        sage: a.display(eV)
        a = (-3/4*u - 3/4*v) du∧dv

    The module `\Omega^1(M)` is nothing but the dual of `\mathfrak{X}(M)`
    (the module of vector fields on `M`)::

        sage: L1 = M.diff_form_module(1) ; L1
        Module Omega^1(M) of 1-forms on the 2-dimensional differentiable
         manifold M
        sage: L1 is XM.dual()
        True

    Since any tensor field of type `(0,1)` is a 1-form, there is a coercion
    map from the set `T^{(0,1)}(M)` of such tensors to `\Omega^1(M)`::

        sage: T01 = M.tensor_field_module((0,1)) ; T01
        Module T^(0,1)(M) of type-(0,1) tensors fields on the 2-dimensional
         differentiable manifold M
        sage: L1.has_coerce_map_from(T01)
        True

    There is also a coercion map in the reverse direction::

        sage: T01.has_coerce_map_from(L1)
        True

    For a degree `p \geq 2`, the coercion holds only in the direction
    `\Omega^p(M)\rightarrow T^{(0,p)}(M)`::

        sage: T02 = M.tensor_field_module((0,2)) ; T02
        Module T^(0,2)(M) of type-(0,2) tensors fields on the 2-dimensional
         differentiable manifold M
        sage: T02.has_coerce_map_from(A)
        True
        sage: A.has_coerce_map_from(T02)
        False

    The coercion map `T^{(0,1)}(M) \rightarrow \Omega^1(M)` in action::

        sage: b = T01([y,x], frame=eU, name='b') ; b
        Tensor field b of type (0,1) on the 2-dimensional differentiable
         manifold M
        sage: b.add_comp_by_continuation(eV, W, c_uv)
        sage: b.display(eU)
        b = y dx + x dy
        sage: b.display(eV)
        b = 1/2*u du - 1/2*v dv
        sage: lb = L1(b) ; lb
        1-form b on the 2-dimensional differentiable manifold M
        sage: lb.display(eU)
        b = y dx + x dy
        sage: lb.display(eV)
        b = 1/2*u du - 1/2*v dv

    The coercion map `\Omega^1(M) \rightarrow T^{(0,1)}(M)` in action::

        sage: tlb = T01(lb) ; tlb
        Tensor field b of type (0,1) on the 2-dimensional differentiable
         manifold M
        sage: tlb.display(eU)
        b = y dx + x dy
        sage: tlb.display(eV)
        b = 1/2*u du - 1/2*v dv
        sage: tlb == b
        True

    The coercion map `\Omega^2(M) \rightarrow T^{(0,2)}(M)` in action::

        sage: ta = T02(a) ; ta
        Tensor field a of type (0,2) on the 2-dimensional differentiable
         manifold M
        sage: ta.display(eU)
        a = 3*x dx⊗dy - 3*x dy⊗dx
        sage: a.display(eU)
        a = 3*x dx∧dy
        sage: ta.display(eV)
        a = (-3/4*u - 3/4*v) du⊗dv + (3/4*u + 3/4*v) dv⊗du
        sage: a.display(eV)
        a = (-3/4*u - 3/4*v) du∧dv

    There is also coercion to subdomains, which is nothing but the restriction
    of the differential form to some subset of its domain::

        sage: L2U = U.diff_form_module(2) ; L2U
        Free module Omega^2(U) of 2-forms on the Open subset U of the
         2-dimensional differentiable manifold M
        sage: L2U.has_coerce_map_from(A)
        True
        sage: a_U = L2U(a) ; a_U
        2-form a on the Open subset U of the 2-dimensional differentiable
         manifold M
        sage: a_U.display(eU)
        a = 3*x dx∧dy
    """
    Element = DiffForm

    def __init__(self, vector_field_module, degree):
        r"""
        Construction a module of differential forms.

        TESTS:

        Module of 2-forms on a non-parallelizable 2-dimensional manifold::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U') ; V = M.open_subset('V')
            sage: M.declare_union(U,V)   # M is the union of U and V
            sage: c_xy.<x,y> = U.chart() ; c_uv.<u,v> = V.chart()
            sage: transf = c_xy.transition_map(c_uv, (x+y, x-y),
            ....:                    intersection_name='W', restrictions1= x>0,
            ....:                    restrictions2= u+v>0)
            sage: inv = transf.inverse()
            sage: from sage.manifolds.differentiable.diff_form_module import \
            ....:                                                DiffFormModule
            sage: A = DiffFormModule(M.vector_field_module(), 2) ; A
            Module Omega^2(M) of 2-forms on the 2-dimensional differentiable
             manifold M
            sage: TestSuite(A).run(skip='_test_elements')

        In the above test suite, ``_test_elements`` is skipped because of the
        ``_test_pickling`` error of the elements (to be fixed in
        :class:`sage.manifolds.differentiable.tensorfield.TensorField`)
        """
        domain = vector_field_module._domain
        dest_map = vector_field_module._dest_map
        name = "Omega^{}(".format(degree) + domain._name
        latex_name = r"\Omega^{{{}}}\left({}".format(degree, domain._latex_name)
        if dest_map is not domain.identity_map():
            dm_name = dest_map._name
            dm_latex_name = dest_map._latex_name
            if dm_name is None:
                dm_name = "unnamed map"
            if dm_latex_name is None:
                dm_latex_name = r"\mathrm{unnamed\; map}"
            name += "," + dm_name
            latex_name += "," + dm_latex_name
        self._name = name + ")"
        self._latex_name = latex_name + r"\right)"
        self._vmodule = vector_field_module
        self._degree = degree
        # the member self._ring is created for efficiency (to avoid calls to
        # self.base_ring()):
        self._ring = domain.scalar_field_algebra()
        Parent.__init__(self, base=self._ring, category=Modules(self._ring))
        self._domain = domain
        self._dest_map = dest_map
        self._ambient_domain = vector_field_module._ambient_domain
        # NB: self._zero_element is not constructed here, since no element
        # can be constructed here, to avoid some infinite recursion.

    #### Parent methods

    def _element_constructor_(self, comp=[], frame=None, name=None,
                              latex_name=None):
        r"""
        Construct a differential form.

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U'); V = M.open_subset('V')
            sage: c_xy.<x,y> = U.chart(); c_uv.<u,v> = V.chart()
            sage: M.declare_union(U,V)
            sage: A = M.diff_form_module(2)
            sage: a = A([[0, x*y], [-x*y, 0]], name='a'); a
            2-form a on the 2-dimensional differentiable manifold M
            sage: a.display(c_xy.frame())
            a = x*y dx∧dy
            sage: A(0) is A.zero()
            True
        """
        try:
            if comp.is_trivial_zero():
                return self.zero()
        except AttributeError:
            if comp == 0:
                return self.zero()
        if isinstance(comp, (DiffForm, DiffFormParal)):
            # coercion by domain restriction
            if (self._degree == comp._tensor_type[1]
                   and self._domain.is_subset(comp._domain)
                   and self._ambient_domain.is_subset(comp._ambient_domain)):
                return comp.restrict(self._domain)
            raise TypeError("cannot convert the {} ".format(comp) +
                            "to an element of {}".format(self))
        if isinstance(comp, TensorField):
            # coercion of a tensor of type (0,1) to a linear form
            tensor = comp # for readability
            if (tensor.tensor_type() == (0,1) and self._degree == 1
                     and tensor._vmodule is self._vmodule):
                resu = self.element_class(self._vmodule, 1, name=tensor._name,
                                          latex_name=tensor._latex_name)
                for dom, rst in tensor._restrictions.items():
                    resu._restrictions[dom] = dom.diff_form_module(1)(rst)
                return resu
            raise TypeError("cannot convert the {} ".format(tensor) +
                            "to an element of {}".format(self))
        if not isinstance(comp, (list, tuple)):
            raise TypeError("cannot convert the {} ".format(comp) +
                            "to an element of {}".format(self))
        # standard construction
        resu = self.element_class(self._vmodule, self._degree, name=name,
                                  latex_name=latex_name)
        if comp:
            resu.set_comp(frame)[:] = comp
        return resu

    def _an_element_(self):
        r"""
        Construct some (unnamed) differential form.

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: U = M.open_subset('U'); V = M.open_subset('V')
            sage: c_xy.<x,y> = U.chart(); c_uv.<u,v> = V.chart()
            sage: M.declare_union(U,V)
            sage: A = M.diff_form_module(2)
            sage: A._an_element_()
            2-form on the 2-dimensional differentiable manifold M
        """
        resu = self.element_class(self._vmodule, self._degree)
        for oc in self._domain.open_covers(trivial=False):
            # the first non-trivial open cover is selected
            for dom in oc:
                vmodule_dom = dom.vector_field_module(
                                         dest_map=self._dest_map.restrict(dom))
                dmodule_dom = vmodule_dom.dual_exterior_power(self._degree)
                resu.set_restriction(dmodule_dom._an_element_())
            return resu
        return resu

    def _coerce_map_from_(self, other):
        r"""
        Determine whether coercion to ``self`` exists from other parent.

        TESTS::

            sage: M = Manifold(3, 'M')
            sage: A1 = M.diff_form_module(1)
            sage: A1._coerce_map_from_(M.tensor_field_module((0,1)))
            True
            sage: A2 = M.diff_form_module(2)
            sage: A2._coerce_map_from_(M.tensor_field_module((0,2)))
            False
            sage: U = M.open_subset('U')
            sage: A2U = U.diff_form_module(2)
            sage: A2U._coerce_map_from_(A2)
            True
            sage: A2._coerce_map_from_(A2U)
            False
        """
        if isinstance(other, (DiffFormModule, DiffFormFreeModule)):
            # coercion by domain restriction
            return (self._degree == other._degree
                    and self._domain.is_subset(other._domain)
                    and self._ambient_domain.is_subset(other._ambient_domain))

        from sage.manifolds.differentiable.tensorfield_module import TensorFieldModule
        if isinstance(other, TensorFieldModule):
            # coercion of a type-(0,1) tensor to a linear form
            return (self._vmodule is other._vmodule and self._degree == 1
                    and other.tensor_type() == (0,1))

        return False

    @cached_method
    def zero(self):
        """
        Return the zero of ``self``.

        EXAMPLES::

            sage: M = Manifold(3, 'M')
            sage: A2 = M.diff_form_module(2)
            sage: A2.zero()
            2-form zero on the 3-dimensional differentiable manifold M
        """
        zero = self._element_constructor_(name='zero', latex_name='0')
        for frame in self._domain._frames:
            if self._dest_map.restrict(frame._domain) == frame._dest_map:
                zero.add_comp(frame)
                # (since new components are initialized to zero)
        zero._is_zero = True  # This element is certainly zero
        zero.set_immutable()
        return zero

    #### End of Parent methods

    def _repr_(self):
        r"""
        Return a string representation of the object.

        TESTS::

            sage: M = Manifold(3, 'M')
            sage: A2 = M.diff_form_module(2)
            sage: A2
            Module Omega^2(M) of 2-forms on
             the 3-dimensional differentiable manifold M
        """
        description = "Module "
        if self._name is not None:
            description += self._name + " "
        description += "of {}-forms ".format(self._degree)
        if self._dest_map is self._domain.identity_map():
            description += "on the {}".format(self._domain)
        else:
            description += "along the {} mapped into the {}".format(
                                            self._domain, self._ambient_domain)
        return description

    def _latex_(self):
        r"""
        Return a LaTeX representation of the object.

        TESTS::

            sage: M = Manifold(3, 'M', latex_name=r'\mathcal{M}')
            sage: A2 = M.diff_form_module(2)
            sage: A2._latex_()
            '\\Omega^{2}\\left(\\mathcal{M}\\right)'
            sage: latex(A2)  # indirect doctest
            \Omega^{2}\left(\mathcal{M}\right)
        """
        if self._latex_name is None:
            return r'\text{' + str(self) + r'}'
        return self._latex_name

    def base_module(self):
        r"""
        Return the vector field module on which the differential form module
        ``self`` is constructed.

        OUTPUT:

        - a
          :class:`~sage.manifolds.differentiable.vectorfield_module.VectorFieldModule`
          representing the module on which ``self`` is defined

        EXAMPLES::

            sage: M = Manifold(3, 'M')
            sage: A2 = M.diff_form_module(2) ; A2
            Module Omega^2(M) of 2-forms on the 3-dimensional differentiable
             manifold M
            sage: A2.base_module()
            Module X(M) of vector fields on the 3-dimensional differentiable
             manifold M
            sage: A2.base_module() is M.vector_field_module()
            True
            sage: U = M.open_subset('U')
            sage: A2U = U.diff_form_module(2) ; A2U
            Module Omega^2(U) of 2-forms on the Open subset U of the
             3-dimensional differentiable manifold M
            sage: A2U.base_module()
            Module X(U) of vector fields on the Open subset U of the
             3-dimensional differentiable manifold M
        """
        return self._vmodule

    tensor = tensor_product = ReflexiveModule_abstract.tensor_product

    def tensor_type(self):
        r"""
        Return the tensor type of ``self`` if ``self`` is a module of 1-forms.

        In this case, the pair `(0, 1)` is returned, indicating that the module
        is identified with the dual of the base module.

        For differential forms of other degrees, an exception is raised.

        EXAMPLES::

            sage: M = Manifold(3, 'M')
            sage: M.diff_form_module(1).tensor_type()
            (0, 1)
            sage: M.diff_form_module(2).tensor_type()
            Traceback (most recent call last):
            ...
            NotImplementedError
        """
        if self._degree == 1:
            return (0, 1)
        raise NotImplementedError

    def degree(self):
        r"""
        Return the degree of the differential forms in ``self``.

        OUTPUT: integer `p` such that ``self`` is a set of `p`-forms

        EXAMPLES::

            sage: M = Manifold(3, 'M')
            sage: M.diff_form_module(1).degree()
            1
            sage: M.diff_form_module(2).degree()
            2
            sage: M.diff_form_module(3).degree()
            3
        """
        return self._degree

# *****************************************************************************


class DiffFormFreeModule(ExtPowerDualFreeModule):
    r"""
    Free module of differential forms of a given degree `p` (`p`-forms) along
    a differentiable manifold `U` with values on a parallelizable manifold `M`.

    Given a differentiable manifold `U` and a differentiable map
    `\Phi:\; U \rightarrow M` to a parallelizable manifold `M` of dimension
    `n`, the set `\Omega^p(U, \Phi)` of `p`-forms along `U` with values on `M`
    is a free module of rank `\binom{n}{p}` over `C^k(U)`, the commutative
    algebra of differentiable scalar fields on `U` (see
    :class:`~sage.manifolds.differentiable.scalarfield_algebra.DiffScalarFieldAlgebra`).
    The standard case of `p`-forms *on* a differentiable manifold `M`
    corresponds to `U = M` and `\Phi = \mathrm{Id}_M`. Other common cases are
    `\Phi` being an immersion and `\Phi` being a curve in `M` (`U` is then an
    open interval of `\RR`).

    .. NOTE::

        This class implements `\Omega^p(U, \Phi)` in the case where `M` is
        parallelizable; `\Omega^p(U, \Phi)` is then a *free* module. If `M`
        is not parallelizable, the class :class:`DiffFormModule` must be used
        instead.

    For the special case of 1-forms, use the class :class:`VectorFieldDualFreeModule`.

    INPUT:

    - ``vector_field_module`` -- free module `\mathfrak{X}(U,\Phi)` of vector
      fields along `U` associated with the map `\Phi: U \rightarrow V`
    - ``degree`` -- positive integer; the degree `p` of the differential forms

    EXAMPLES:

    Free module of 2-forms on a parallelizable 3-dimensional manifold::

        sage: M = Manifold(3, 'M')
        sage: X.<x,y,z> = M.chart()
        sage: XM = M.vector_field_module() ; XM
        Free module X(M) of vector fields on the 3-dimensional differentiable
         manifold M
        sage: A = M.diff_form_module(2) ; A
        Free module Omega^2(M) of 2-forms on the 3-dimensional differentiable
         manifold M
        sage: latex(A)
        \Omega^{2}\left(M\right)

    ``A`` is nothing but the second exterior power of the dual of ``XM``, i.e.
    we have `\Omega^{2}(M) = \Lambda^2(\mathfrak{X}(M)^*)` (see
    :class:`~sage.tensor.modules.ext_pow_free_module.ExtPowerDualFreeModule`)::

        sage: A is XM.dual_exterior_power(2)
        True

    `\Omega^{2}(M)` is a module over the algebra `C^k(M)` of (differentiable)
    scalar fields on `M`::

        sage: A.category()
        Category of finite dimensional modules over Algebra of differentiable
         scalar fields on the 3-dimensional differentiable manifold M
        sage: CM = M.scalar_field_algebra() ; CM
        Algebra of differentiable scalar fields on the 3-dimensional
         differentiable manifold M
        sage: A in Modules(CM)
        True
        sage: A.base_ring()
        Algebra of differentiable scalar fields on
         the 3-dimensional differentiable manifold M
        sage: A.base_module()
        Free module X(M) of vector fields on
         the 3-dimensional differentiable manifold M
        sage: A.base_module() is XM
        True
        sage: A.rank()
        3

    Elements can be constructed from `A`. In particular, ``0`` yields
    the zero element of `A`::

        sage: A(0)
        2-form zero on the 3-dimensional differentiable manifold M
        sage: A(0) is A.zero()
        True

    while nonzero elements are constructed by providing their components
    in a given vector frame::

        sage: comp = [[0,3*x,-z],[-3*x,0,4],[z,-4,0]]
        sage: a = A(comp, frame=X.frame(), name='a') ; a
        2-form a on the 3-dimensional differentiable manifold M
        sage: a.display()
        a = 3*x dx∧dy - z dx∧dz + 4 dy∧dz

    An alternative is to construct the 2-form from an empty list of
    components and to set the nonzero nonredundant components afterwards::

        sage: a = A([], name='a')
        sage: a[0,1] = 3*x  # component in the manifold's default frame
        sage: a[0,2] = -z
        sage: a[1,2] = 4
        sage: a.display()
        a = 3*x dx∧dy - z dx∧dz + 4 dy∧dz

    The module `\Omega^1(M)` is nothing but the dual of `\mathfrak{X}(M)`
    (the free module of vector fields on `M`)::

        sage: L1 = M.diff_form_module(1) ; L1
        Free module Omega^1(M) of 1-forms on the 3-dimensional differentiable
         manifold M
        sage: L1 is XM.dual()
        True

    Since any tensor field of type `(0,1)` is a 1-form, it is also equal to
    the set `T^{(0,1)}(M)` of such tensors to `\Omega^1(M)`::

        sage: T01 = M.tensor_field_module((0,1)) ; T01
        Free module Omega^1(M) of 1-forms on the 3-dimensional differentiable manifold M
        sage: L1 is T01
        True

    For a degree `p \geq 2`, the coercion holds only in the direction
    `\Omega^p(M) \rightarrow T^{(0,p)}(M)`::

        sage: T02 = M.tensor_field_module((0,2)); T02
        Free module T^(0,2)(M) of type-(0,2) tensors fields on the
         3-dimensional differentiable manifold M
        sage: T02.has_coerce_map_from(A)
        True
        sage: A.has_coerce_map_from(T02)
        False

    The coercion map `\Omega^2(M) \rightarrow T^{(0,2)}(M)` in action::

        sage: T02 = M.tensor_field_module((0,2)) ; T02
        Free module T^(0,2)(M) of type-(0,2) tensors fields on the
         3-dimensional differentiable manifold M
        sage: ta = T02(a) ; ta
        Tensor field a of type (0,2) on the 3-dimensional differentiable
         manifold M
        sage: ta.display()
        a = 3*x dx⊗dy - z dx⊗dz - 3*x dy⊗dx + 4 dy⊗dz + z dz⊗dx - 4 dz⊗dy
        sage: a.display()
        a = 3*x dx∧dy - z dx∧dz + 4 dy∧dz
        sage: ta.symmetries()  # the antisymmetry is preserved
        no symmetry;  antisymmetry: (0, 1)

    There is also coercion to subdomains, which is nothing but the
    restriction of the differential form to some subset of its domain::

        sage: U = M.open_subset('U', coord_def={X: x^2+y^2<1})
        sage: B = U.diff_form_module(2) ; B
        Free module Omega^2(U) of 2-forms on the Open subset U of the
         3-dimensional differentiable manifold M
        sage: B.has_coerce_map_from(A)
        True
        sage: a_U = B(a) ; a_U
        2-form a on the Open subset U of the 3-dimensional differentiable
         manifold M
        sage: a_U.display()
        a = 3*x dx∧dy - z dx∧dz + 4 dy∧dz
    """

    Element = DiffFormParal

    def __init__(self, vector_field_module, degree):
        r"""
        Construct a free module of differential forms.

        TESTS::

            sage: M = Manifold(3, 'M')
            sage: X.<x,y,z> = M.chart()
            sage: from sage.manifolds.differentiable.diff_form_module import DiffFormFreeModule
            sage: A = DiffFormFreeModule(M.vector_field_module(), 2) ; A
            Free module Omega^2(M) of 2-forms on
             the 3-dimensional differentiable manifold M
            sage: TestSuite(A).run()
        """
        domain = vector_field_module._domain
        dest_map = vector_field_module._dest_map
        name = "Omega^{}(".format(degree) + domain._name
        latex_name = r"\Omega^{{{}}}\left({}".format(degree, domain._latex_name)
        if dest_map is not domain.identity_map():
            dm_name = dest_map._name
            dm_latex_name = dest_map._latex_name
            if dm_name is None:
                dm_name = "unnamed map"
            if dm_latex_name is None:
                dm_latex_name = r"\mathrm{unnamed\; map}"
            name += "," + dm_name
            latex_name += "," + dm_latex_name
        name += ")"
        latex_name += r"\right)"
        ExtPowerDualFreeModule.__init__(self, vector_field_module, degree,
                                        name=name, latex_name=latex_name)
        self._domain = domain
        self._dest_map = dest_map
        self._ambient_domain = vector_field_module._ambient_domain

    #### Parent methods

    def _element_constructor_(self, comp=[], frame=None, name=None,
                              latex_name=None):
        r"""
        Construct a differential form.

        TESTS::

            sage: M = Manifold(2, 'M')
            sage: X.<x,y> = M.chart()  # makes M parallelizable
            sage: A = M.diff_form_module(2)
            sage: a = A([[0, x], [-x, 0]], name='a'); a
            2-form a on the 2-dimensional differentiable manifold M
            sage: a.display()
            a = x dx∧dy
            sage: A(0) is A.zero()
            True

        Check that :issue:`27658` is fixed::

            sage: f = M.scalar_field(x)
            sage: f in A
            False
        """
        try:
            if comp.is_trivial_zero():
                return self.zero()
        except AttributeError:
            if comp == 0:
                return self.zero()
        if isinstance(comp, (DiffForm, DiffFormParal)):
            # coercion by domain restriction
            if (self._degree == comp._tensor_type[1]
                    and self._domain.is_subset(comp._domain)
                    and self._ambient_domain.is_subset(comp._ambient_domain)):
                return comp.restrict(self._domain)
            raise TypeError("cannot convert the {} ".format(comp) +
                            "to a differential form in {}".format(self))
        if isinstance(comp, TensorFieldParal):
            # coercion of a tensor of type (0,1) to a linear form
            tensor = comp # for readability
            if (tensor.tensor_type() == (0,1) and self._degree == 1
                     and tensor._fmodule is self._fmodule):
                resu = self.element_class(self._fmodule, 1, name=tensor._name,
                                          latex_name=tensor._latex_name)
                for frame, comp in tensor._components.items():
                    resu._components[frame] = comp.copy()
                return resu
            raise TypeError("cannot convert the {} ".format(tensor) +
                            "to an element of {}".format(self))
        if not isinstance(comp, (list, tuple)):
            raise TypeError("cannot convert the {} ".format(comp) +
                            "to an element of {}".format(self))
        # standard construction
        resu = self.element_class(self._fmodule, self._degree, name=name,
                                  latex_name=latex_name)
        if comp:
            resu.set_comp(frame)[:] = comp
        return resu

    # Rem: _an_element_ is declared in the superclass ExtPowerDualFreeModule

    def _coerce_map_from_(self, other):
        r"""
        Determine whether coercion to ``self`` exists from other parent.

        TESTS::

            sage: M = Manifold(3, 'M')
            sage: X.<x,y,z> = M.chart()
            sage: A2 = M.diff_form_module(2)
            sage: U = M.open_subset('U', coord_def = {X: z<0})
            sage: A2U = U.diff_form_module(2)
            sage: A2U._coerce_map_from_(A2)
            True
            sage: A2._coerce_map_from_(A2U)
            False
            sage: A1 = M.diff_form_module(1)
            sage: A2U._coerce_map_from_(A1)
            False
            sage: A1._coerce_map_from_(M.tensor_field_module((0,1)))
            True
            sage: A1._coerce_map_from_(M.tensor_field_module((1,0)))
            False
        """
        if isinstance(other, (DiffFormModule, DiffFormFreeModule)):
            # coercion by domain restriction
            return (self._degree == other._degree
                    and self._domain.is_subset(other._domain)
                    and self._ambient_domain.is_subset(other._ambient_domain))

        from sage.manifolds.differentiable.tensorfield_module import (
            TensorFieldFreeModule,
        )
        if isinstance(other, TensorFieldFreeModule):
            # coercion of a type-(0,1) tensor to a linear form
            return (self._fmodule is other._fmodule and self._degree == 1
                    and other.tensor_type() == (0,1))
        return False

    #### End of Parent methods

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        TESTS::

            sage: M = Manifold(3, 'M')
            sage: X.<x,y,z> = M.chart()
            sage: A = M.diff_form_module(2)
            sage: A
            Free module Omega^2(M) of 2-forms on
             the 3-dimensional differentiable manifold M
        """
        description = "Free module "
        if self._name is not None:
            description += self._name + " "
        description += "of {}-forms ".format(self._degree)
        if self._dest_map is self._domain.identity_map():
            description += "on the {}".format(self._domain)
        else:
            description += "along the {} mapped into the {}".format(
                                            self._domain, self._ambient_domain)
        return description


class VectorFieldDualFreeModule(DiffFormFreeModule):
    r"""
    Free module of differential 1-forms along a differentiable manifold `U`
    with values on a parallelizable manifold `M`.

    Given a differentiable manifold `U` and a differentiable map
    `\Phi:\; U \rightarrow M` to a parallelizable manifold `M` of dimension
    `n`, the set `\Omega^1(U, \Phi)` of 1-forms along `U` with values on `M`
    is a free module of rank `n` over `C^k(U)`, the commutative
    algebra of differentiable scalar fields on `U` (see
    :class:`~sage.manifolds.differentiable.scalarfield_algebra.DiffScalarFieldAlgebra`).
    The standard case of 1-forms *on* a differentiable manifold `M`
    corresponds to `U = M` and `\Phi = \mathrm{Id}_M`. Other common cases are
    `\Phi` being an immersion and `\Phi` being a curve in `M` (`U` is then an
    open interval of `\RR`).

    .. NOTE::

        This class implements `\Omega^1(U, \Phi)` in the case where `M` is
        parallelizable; `\Omega^1(U, \Phi)` is then a *free* module. If `M`
        is not parallelizable, the class :class:`DiffFormModule` must be used
        instead.

    INPUT:

    - ``vector_field_module`` -- free module `\mathfrak{X}(U,\Phi)` of vector
      fields along `U` associated with the map `\Phi: U \rightarrow V`

    EXAMPLES:

    Free module of 1-forms on a parallelizable 3-dimensional manifold::

        sage: M = Manifold(3, 'M')
        sage: X.<x,y,z> = M.chart()
        sage: XM = M.vector_field_module() ; XM
        Free module X(M) of vector fields on the 3-dimensional differentiable
         manifold M
        sage: A = M.diff_form_module(1) ; A
        Free module Omega^1(M) of 1-forms on the 3-dimensional differentiable manifold M
        sage: latex(A)
        \Omega^{1}\left(M\right)

    ``A`` is nothing but the dual of ``XM`` (the free module of vector fields on `M`)
    and thus also equal to the 1st exterior
    power of the dual, i.e. we have `\Omega^{1}(M) = \Lambda^1(\mathfrak{X}(M)^*)
    = \mathfrak{X}(M)^*` (See
    :class:`~sage.tensor.modules.ext_pow_free_module.ExtPowerDualFreeModule`)::

        sage: A is XM.dual_exterior_power(1)
        True

    `\Omega^{1}(M)` is a module over the algebra `C^k(M)` of (differentiable)
    scalar fields on `M`::

        sage: A.category()
        Category of finite dimensional modules over Algebra of differentiable
         scalar fields on the 3-dimensional differentiable manifold M
        sage: CM = M.scalar_field_algebra() ; CM
        Algebra of differentiable scalar fields on the 3-dimensional
         differentiable manifold M
        sage: A in Modules(CM)
        True
        sage: A.base_ring()
        Algebra of differentiable scalar fields on
         the 3-dimensional differentiable manifold M
        sage: A.base_module()
        Free module X(M) of vector fields on
         the 3-dimensional differentiable manifold M
        sage: A.base_module() is XM
        True
        sage: A.rank()
        3

    Elements can be constructed from `A`. In particular, ``0`` yields
    the zero element of `A`::

        sage: A(0)
        1-form zero on the 3-dimensional differentiable manifold M
        sage: A(0) is A.zero()
        True

    while nonzero elements are constructed by providing their components
    in a given vector frame::

        sage: comp = [3*x,-z,4]
        sage: a = A(comp, frame=X.frame(), name='a') ; a
        1-form a on the 3-dimensional differentiable manifold M
        sage: a.display()
        a = 3*x dx - z dy + 4 dz

    An alternative is to construct the 1-form from an empty list of
    components and to set the nonzero nonredundant components afterwards::

        sage: a = A([], name='a')
        sage: a[0] = 3*x  # component in the manifold's default frame
        sage: a[1] = -z
        sage: a[2] = 4
        sage: a.display()
        a = 3*x dx - z dy + 4 dz

    Since any tensor field of type `(0,1)` is a 1-form, there is a coercion
    map from the set `T^{(0,1)}(M)` of such tensors to `\Omega^1(M)`::

        sage: T01 = M.tensor_field_module((0,1)) ; T01
        Free module Omega^1(M) of 1-forms on the 3-dimensional differentiable manifold M
        sage: A.has_coerce_map_from(T01)
        True

    There is also a coercion map in the reverse direction::

        sage: T01.has_coerce_map_from(A)
        True

    The coercion map `T^{(0,1)}(M) \rightarrow \Omega^1(M)` in action::

        sage: b = T01([-x,2,3*y], name='b'); b
        1-form b on the 3-dimensional differentiable manifold M
        sage: b.display()
        b = -x dx + 2 dy + 3*y dz
        sage: lb = A(b) ; lb
        1-form b on the 3-dimensional differentiable manifold M
        sage: lb.display()
        b = -x dx + 2 dy + 3*y dz

    The coercion map `\Omega^1(M) \rightarrow T^{(0,1)}(M)` in action::

        sage: tlb = T01(lb); tlb
        1-form b on the 3-dimensional differentiable manifold M
        sage: tlb == b
        True
    """

    def __init__(self, vector_field_module):
        r"""
        Construct a free module of differential 1-forms.

        TESTS::

            sage: M = Manifold(3, 'M')
            sage: X.<x,y,z> = M.chart()
            sage: A = M.vector_field_module().dual(); A
            Free module Omega^1(M) of 1-forms on the 3-dimensional differentiable manifold M
            sage: TestSuite(A).run()
        """
        DiffFormFreeModule.__init__(self, vector_field_module, 1)

    def tensor_type(self):
        r"""
        Return the tensor type of ``self``.

        EXAMPLES::

            sage: M = Manifold(3, 'M')
            sage: X.<x,y,z> = M.chart()
            sage: A = M.vector_field_module().dual(); A
            Free module Omega^1(M) of 1-forms on the 3-dimensional differentiable manifold M
            sage: A.tensor_type()
            (0, 1)
        """
        return (0, 1)



# sage.doctest: needs sage.rings.finite_rings
r"""
Unitary Groups `GU(n,q)` and `SU(n,q)` with GAP
"""

# ****************************************************************************
#       Copyright (C) 2006 David Joyner and William Stein
#                     2018 Sebastian Oehms
#                     2018 Travis Scrimshaw
#                     2023 Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.groups.matrix_gps.finitely_generated_gap import FinitelyGeneratedMatrixGroup_gap
from sage.groups.matrix_gps.named_group_gap import NamedMatrixGroup_gap
from sage.groups.matrix_gps.unitary import UnitaryMatrixGroup_generic
from sage.misc.cachefunc import cached_method


class UnitaryMatrixGroup_gap(UnitaryMatrixGroup_generic, NamedMatrixGroup_gap, FinitelyGeneratedMatrixGroup_gap):
    r"""
    The general or special unitary group in GAP.

    TESTS:

    Check that :issue:`20867` is fixed::

        sage: from sage.groups.matrix_gps.finitely_generated_gap import FinitelyGeneratedMatrixGroup_gap
        sage: G = GU(3,3)
        sage: isinstance(G, FinitelyGeneratedMatrixGroup_gap)
        True
    """

    @cached_method
    def invariant_form(self):
        """
        Return the hermitian form preserved by the unitary group.

        OUTPUT: a square matrix describing the bilinear form

        EXAMPLES::

            sage: G32 = GU(3,2)
            sage: G32.invariant_form()
            [0 0 1]
            [0 1 0]
            [1 0 0]
        """
        d = self.degree()
        R = self.base_ring()
        # note that self.gap().InvariantSesquilinearForm()['matrix'].matrix().base_ring() != R for example for self = GU(3.2)
        # therefore we have to coerce into the right matrix space
        from sage.matrix.constructor import matrix
        m = matrix(R, d, d, self.gap().InvariantSesquilinearForm()['matrix'].matrix())
        m.set_immutable()
        return m



from sage.libs.mpfr.types cimport mpfr_prec_t
from sage.libs.mpfi.types cimport mpfi_t

cimport sage.structure.element
from sage.rings.real_mpfi cimport RealIntervalFieldElement, RealIntervalField_class


cdef class ComplexIntervalFieldElement(sage.structure.element.FieldElement):
    cdef mpfi_t __re
    cdef mpfi_t __im
    cdef mpfr_prec_t _prec
    cdef object _multiplicative_order

    cdef inline ComplexIntervalFieldElement _new(self):
        """
        Quickly create a new complex interval with the same parent as
        ``self``.
        """
        cdef type t = type(self)
        cdef object _multiplicative_order = None
        return t.__new__(t, self._parent)

    cdef inline RealIntervalFieldElement _new_real(self):
        """
        Quickly create a new real interval with the same precision as
        ``self``.
        """
        P = <RealIntervalField_class>(self._parent.real_field())
        return P._new()



"""
Circles
"""
#*****************************************************************************
#       Copyright (C) 2006 Alex Clemesha <clemesha@gmail.com>,
#                          William Stein <wstein@gmail.com>,
#                     2008 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from .primitive import GraphicPrimitive
from sage.misc.decorators import options, rename_keyword
from sage.plot.colors import to_mpl_color
from math import sin, cos, pi


class Circle(GraphicPrimitive):
    """
    Primitive class for the :class:`Circle` graphics type.  See ``circle?`` for information
    about actually plotting circles.

    INPUT:

    - ``x`` -- `x`-coordinate of center of Circle

    - ``y`` -- `y`-coordinate of center of Circle

    - ``r`` -- radius of Circle object

    - ``options`` -- dictionary of valid plot options to pass to constructor

    EXAMPLES:

    Note this should normally be used indirectly via ``circle``::

        sage: from sage.plot.circle import Circle
        sage: C = Circle(2,3,5,{'zorder':2})
        sage: C
        Circle defined by (2.0,3.0) with r=5.0
        sage: C.options()['zorder']
        2
        sage: C.r
        5.0

    TESTS:

    We test creating a circle::

        sage: C = circle((2,3), 5)
    """
    def __init__(self, x, y, r, options):
        """
        Initialize base class Circle.

        EXAMPLES::

            sage: C = circle((2,3), 5, edgecolor='red', alpha=.5, fill=True)
            sage: C[0].x
            2.0
            sage: C[0].r
            5.0
            sage: C[0].options()['edgecolor']
            'red'
            sage: C[0].options()['alpha']
            0.500000000000000
        """
        self.x = float(x)
        self.y = float(y)
        self.r = float(r)
        GraphicPrimitive.__init__(self, options)

    def get_minmax_data(self):
        """
        Return a dictionary with the bounding box data.

        EXAMPLES::

            sage: p = circle((3, 3), 1)
            sage: d = p.get_minmax_data()
            sage: d['xmin']
            2.0
            sage: d['ymin']
            2.0
        """
        from sage.plot.plot import minmax_data
        return minmax_data([self.x - self.r, self.x + self.r],
                           [self.y - self.r, self.y + self.r],
                           dict=True)

    def _allowed_options(self):
        """
        Return the allowed options for the Circle class.

        EXAMPLES::

            sage: p = circle((3, 3), 1)
            sage: p[0]._allowed_options()['alpha']
            'How transparent the figure is.'
            sage: p[0]._allowed_options()['facecolor']
            '2D only: The color of the face as an RGB tuple.'
        """
        return {'alpha': 'How transparent the figure is.',
                'fill': 'Whether or not to fill the circle.',
                'legend_label': 'The label for this item in the legend.',
                'legend_color': 'The color of the legend text.',
                'thickness': 'How thick the border of the circle is.',
                'edgecolor': '2D only: The color of the edge as an RGB tuple.',
                'facecolor': '2D only: The color of the face as an RGB tuple.',
                'rgbcolor': 'The color (edge and face) as an RGB tuple.',
                'hue': 'The color given as a hue.',
                'zorder': '2D only: The layer level in which to draw',
                'linestyle': "2D only: The style of the line, which is one of "
                "'dashed', 'dotted', 'solid', 'dashdot', or '--', ':', '-', '-.', "
                "respectively.",
                'clip': 'Whether or not to clip the circle.'}

    def _repr_(self):
        """
        String representation of Circle primitive.

        EXAMPLES::

            sage: C = circle((2,3), 5)
            sage: c = C[0]; c
            Circle defined by (2.0,3.0) with r=5.0
        """
        return f"Circle defined by ({self.x},{self.y}) with r={self.r}"

    def _render_on_subplot(self, subplot):
        """
        TESTS::

            sage: from math import pi
            sage: C = circle((2,pi), 2, edgecolor='black', facecolor='green', fill=True)
        """
        from matplotlib import patches
        from sage.plot.misc import get_matplotlib_linestyle

        options = self.options()
        p = patches.Circle((float(self.x), float(self.y)), float(self.r), clip_on=options['clip'])
        if not options['clip']:
            self._bbox_extra_artists = [p]
        p.set_linewidth(float(options['thickness']))
        p.set_fill(options['fill'])
        a = float(options['alpha'])
        p.set_alpha(a)
        ec = to_mpl_color(options['edgecolor'])
        fc = to_mpl_color(options['facecolor'])
        if 'rgbcolor' in options:
            ec = fc = to_mpl_color(options['rgbcolor'])
        p.set_edgecolor(ec)
        p.set_facecolor(fc)
        p.set_linestyle(get_matplotlib_linestyle(options['linestyle'],return_type='long'))
        p.set_label(options['legend_label'])
        z = int(options.pop('zorder', 0))
        p.set_zorder(z)
        subplot.add_patch(p)

    def plot3d(self, z=0, **kwds):
        """
        Plots a 2D circle (actually a 50-gon) in 3D,
        with default height zero.

        INPUT:

        - ``z`` -- (optional) 3D height above `xy`-plane

        EXAMPLES::

            sage: circle((0,0), 1).plot3d()
            Graphics3d Object

        This example uses this method implicitly, but does not pass
        the optional parameter z to this method::

            sage: sum(circle((random(),random()), random()).plot3d(z=random())
            ....:     for _ in range(20))
            Graphics3d Object

        .. PLOT::

            P = sum([circle((random(),random()), random()).plot3d(z=random()) for _ in range(20)])
            sphinx_plot(P)

        These examples are explicit, and pass z to this method::

            sage: from math import pi
            sage: C = circle((2,pi), 2, hue=.8, alpha=.3, fill=True)
            sage: c = C[0]
            sage: d = c.plot3d(z=2)
            sage: d.texture.opacity
            0.3

        ::

            sage: C = circle((2,pi), 2, hue=.8, alpha=.3, linestyle='dotted')
            sage: c = C[0]
            sage: d = c.plot3d(z=2)
            sage: d.jmol_repr(d.testing_render_params())[0][-1]
            'color $line_1 translucent 0.7 [204,0,255]'
        """
        options = dict(self.options())
        fill = options['fill']
        for s in ['clip', 'edgecolor', 'facecolor', 'fill', 'linestyle',
                'zorder']:
            if s in options:
                del options[s]

        n = 50
        dt = float(2*pi/n)
        x, y, r = self.x, self.y, self.r
        xdata = [x+r*cos(t*dt) for t in range(n+1)]
        ydata = [y+r*sin(t*dt) for t in range(n+1)]
        if fill:
            from .polygon import Polygon
            return Polygon(xdata, ydata, options).plot3d(z)
        from .line import Line
        return Line(xdata, ydata, options).plot3d().translate((0,0,z))


@rename_keyword(color='rgbcolor')
@options(alpha=1, fill=False, thickness=1, edgecolor='blue', facecolor='blue', linestyle='solid',
         zorder=5, legend_label=None, legend_color=None, clip=True, aspect_ratio=1.0)
def circle(center, radius, **options):
    """
    Return a circle at a point center = `(x,y)` (or `(x,y,z)` and
    parallel to the `xy`-plane) with radius = `r`.  Type
    ``circle.options`` to see all options.

    OPTIONS:

    - ``alpha`` -- (default: 1)

    - ``fill`` -- (default: ``False``)

    - ``thickness`` -- (default: 1)

    - ``linestyle`` -- (default: ``'solid'``) (2D plotting only) the style of the
      line, which is one of ``'dashed'``, ``'dotted'``, ``'solid'``, ``'dashdot'``,
      or ``'--'``, ``':'``, ``'-'``, ``'-.'``, respectively

    - ``edgecolor`` -- (default: ``'blue'``) 2D plotting only

    - ``facecolor`` -- (default: ``'blue'``) 2D plotting only, useful only
      if ``fill=True``

    - ``rgbcolor`` -- 2D or 3D plotting.  This option overrides
      ``edgecolor`` and ``facecolor`` for 2D plotting

    - ``legend_label`` -- the label for this item in the legend

    - ``legend_color`` -- the color for the legend label

    EXAMPLES:

    The default color is blue, the default linestyle is solid, but this is easy to change::

        sage: c = circle((1,1), 1)
        sage: c
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(circle((1,1), 1))

    ::

        sage: c = circle((1,1), 1, rgbcolor=(1,0,0), linestyle='-.')
        sage: c
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        c = circle((1,1), 1, rgbcolor=(1,0,0), linestyle='-.')
        sphinx_plot(c)

    We can also use this command to plot three-dimensional circles parallel
    to the `xy`-plane::

        sage: c = circle((1,1,3), 1, rgbcolor=(1,0,0))
        sage: c
        Graphics3d Object
        sage: type(c)
        <class 'sage.plot.plot3d.base.TransformGroup'>

    .. PLOT::

        c = circle((1,1,3), 1, rgbcolor=(1,0,0))
        sphinx_plot(c)

    To correct the aspect ratio of certain graphics, it is necessary
    to show with a ``figsize`` of square dimensions::

        sage: c.show(figsize=[5,5],xmin=-1,xmax=3,ymin=-1,ymax=3)

    Here we make a more complicated plot, with many circles of different colors::

        sage: g = Graphics()
        sage: step = 6; ocur = 1/5; paths = 16
        sage: PI = math.pi    # numerical for speed -- fine for graphics
        sage: for r in range(1,paths+1):
        ....:     for x,y in [((r+ocur)*math.cos(n), (r+ocur)*math.sin(n))
        ....:                 for n in srange(0, 2*PI+PI/step, PI/step)]:
        ....:         g += circle((x,y), ocur, rgbcolor=hue(r/paths))
        ....:     rnext = (r+1)^2
        ....:     ocur = (rnext-r)-ocur
        sage: g.show(xmin=-(paths+1)^2, xmax=(paths+1)^2,
        ....:        ymin=-(paths+1)^2, ymax=(paths+1)^2, figsize=[6,6])

    .. PLOT::

        g = Graphics()
        step=6; ocur=1/5; paths=16;
        PI = math.pi    # numerical for speed -- fine for graphics
        for r in range(1,paths+1):
             for x,y in [((r+ocur)*math.cos(n), (r+ocur)*math.sin(n)) for n in srange(0, 2*PI+PI/step, PI/step)]:
                 g += circle((x,y), ocur, rgbcolor=hue(r*1.0/paths))
             rnext = (r+1)**2
             ocur = (rnext-r)-ocur
        g.set_axes_range(-(paths+1)**2,(paths+1)**2,-(paths+1)**2,(paths+1)**2)
        sphinx_plot(g)

    Note that the ``rgbcolor`` option overrides the other coloring options.
    This produces red fill in a blue circle::

        sage: circle((2,3), 1, fill=True, edgecolor='blue', facecolor='red')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(circle((2,3), 1, fill=True, edgecolor='blue', facecolor='red'))

    This produces an all-green filled circle::

        sage: circle((2,3), 1, fill=True, edgecolor='blue', rgbcolor='green')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(circle((2,3), 1, fill=True, edgecolor='blue', rgbcolor='green'))

    The option ``hue`` overrides *all* other options, so be careful with its use.
    This produces a purplish filled circle::

        sage: circle((2,3), 1, fill=True, edgecolor='blue', rgbcolor='green', hue=.8)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        C = circle((2,3), 1, fill=True, edgecolor='blue', rgbcolor='green', hue=.8)
        sphinx_plot(C)

    And circles with legends::

        sage: circle((4,5), 1, rgbcolor='yellow', fill=True,
        ....:        legend_label='the sun').show(xmin=0, ymin=0)

    .. PLOT::

        C = circle((4,5), 1, rgbcolor='yellow', fill=True, legend_label='the sun')
        C.set_axes_range(xmin=0, ymin=0)
        sphinx_plot(C)

    ::

        sage: circle((4,5), 1,
        ....:        legend_label='the sun', legend_color='yellow').show(xmin=0, ymin=0)

    .. PLOT::

        C = circle((4,5), 1, legend_label='the sun', legend_color='yellow')
        C.set_axes_range(xmin=0, ymin=0)
        sphinx_plot(C)

    Extra options will get passed on to show(), as long as they are valid::

        sage: circle((0, 0), 2, figsize=[10,10])  # That circle is huge!
        Graphics object consisting of 1 graphics primitive

    ::

        sage: circle((0, 0), 2).show(figsize=[10,10])  # These are equivalent

    TESTS:

    We cannot currently plot circles in more than three dimensions::

        sage: circle((1,1,1,1), 1, rgbcolor=(1,0,0))
        Traceback (most recent call last):
        ...
        ValueError: the center of a plotted circle should have two or three coordinates

    The default aspect ratio for a circle is 1.0::

        sage: P = circle((1,1), 1)
        sage: P.aspect_ratio()
        1.0

    Verify that :issue:`36153` does not arise::

        sage: C = circle((1,1), 1, legend_label='test')
    """
    from sage.plot.graphics import Graphics

    # Reset aspect_ratio to 'automatic' in case scale is 'semilog[xy]'.
    # Otherwise matplotlib complains.
    scale = options.get('scale', None)
    if isinstance(scale, (list, tuple)):
        scale = scale[0]
    if scale == 'semilogy' or scale == 'semilogx':
        options['aspect_ratio'] = 'automatic'

    g = Graphics()
    g._set_extra_kwds(Graphics._extract_kwds_for_show(options))
    g.add_primitive(Circle(center[0], center[1], radius, options))
    if options['legend_label']:
        g.legend(True)
        g._legend_colors = [options['legend_color']]
    if len(center) == 2:
        return g
    if len(center) == 3:
        return g[0].plot3d(z=center[2])
    raise ValueError('the center of a plotted circle should have '
                     'two or three coordinates')



"""
Dense matrices over multivariate polynomials over fields.

AUTHOR:

- Enrique Artal (2023-??): initial version
"""

# *****************************************************************************
#       Copyright (C) 2023 Enrique Artal <artal@unizar.es>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************
from sage.matrix.constructor import identity_matrix


cdef class Matrix_laurent_mpolynomial_dense(Matrix_generic_dense):
    """
    Dense matrix over a Laurent multivariate polynomial ring over a field.
    """
    def laurent_matrix_reduction(self):
        """
        From a matrix ``self`` of Laurent polynomials, apply elementary operations
        to obtain a matrix ``P`` of polynomials such that the variables do not divide
        any column and any row.

        OUTPUT:

        Three matrices ``L``, ``P``, ``R`` such that ``self`` equals ``L P R``,
        where ``L`` and ``R`` are diagonal with monomial entries.

        EXAMPLES:

            sage: R.<x, y> = LaurentPolynomialRing(QQ)
            sage: L = [1/3*x^-1*y - 6*x^-2*y^2 - 1/2*x^-2*y, 1/5*x + 1/2*y + 1/6]
            sage: L += [1/2 - 5*x^-1*y - 2*x^-1, -1/3*y^-2 - 4*x^-1*y^-1 + 11*x^-1*y^-2]
            sage: A = matrix(R, 2, L)
            sage: lf, P, rg = A.laurent_matrix_reduction()
            sage: lf
            [     x^-2         0]
            [        0 x^-1*y^-2]
            sage: P
            [            1/3*x - 6*y - 1/2 1/5*x^3 + 1/2*x^2*y + 1/6*x^2]
            [        1/2*x*y - 5*y^2 - 2*y             -1/3*x - 4*y + 11]
            sage: rg
            [y 0]
            [0 1]
        """
        R = self.base_ring()
        n_rows, n_cols = self.dimensions()
        mat_l = identity_matrix(R, n_rows)
        mat_r = identity_matrix(R, n_cols)
        res = self.__copy__()
        for j, rw in enumerate(res.rows()):
            for t in R.gens():
                n = min(mon.degree(t) for a in rw for _, mon in a)
                res.rescale_row(j, t ** -n)
                mat_l.rescale_col(j, t ** n)
        for j, cl in enumerate(res.columns()):
            for t in R.gens():
                n = min(mon.degree(t) for a in cl for _, mon in a)
                res.rescale_col(j, t ** -n)
                mat_r.rescale_row(j, t ** n)
        res = res.change_ring(R.polynomial_ring())
        return mat_l, res, mat_r

    def _fitting_ideal(self, i):
        r"""
        Return the `i`-th Fitting ideal of the matrix.

        This is the ideal generated
        by the `n - i` minors, where `n` is the number of columns.

        INPUT:

        - ``i`` -- integer

        OUTPUT: an ideal on the base ring

        EXAMPLES::

            sage: R.<x,y,z> = LaurentPolynomialRing(QQ)
            sage: M = matrix(R, [[2*x^-1-z, 0, y-z^-2, 0], [0, z - y^-1, z - x, 0],[z - y, x^-2 - y, 0, z]])
            sage: M
            [-z + 2*x^-1           0    y - z^-2           0]
            [          0    z - y^-1      -x + z           0]
            [     -y + z   -y + x^-2           0           z]
            sage: M.fitting_ideal(0)
            Ideal (0) of Multivariate Laurent Polynomial Ring in x, y, z over Rational Field
            sage: M.fitting_ideal(1) == M._fitting_ideal(1)
            True
            sage: M.fitting_ideal(1).groebner_basis()
            (x^4 - 2*x^3*y - x*z^3 - 4*x^2*y + 8*x*y^2 + 4*x*y*z + 2*z^2 - 8*y,
            x*y*z^2 - x*z - 2*y*z + 2,
            x^2*z - x*z^2 - 2*x + 2*z,
            y^2*z + 1/4*x^2 - 1/2*x*y - 1/4*x*z - y + 1/2)
            sage: M.fitting_ideal(2).groebner_basis()
            (1,)
            sage: M.fitting_ideal(3).groebner_basis()
            (1,)
            sage: M.fitting_ideal(4).groebner_basis()
            (1,)
            sage: [R.ideal(M.minors(i)) == M._fitting_ideal(4 - i) for i in range(5)]
            [True, True, True, True, True]
        """
        R = self.base_ring()
        S = R.polynomial_ring()
        A = self.laurent_matrix_reduction()[1].change_ring(S)
        J = A._fitting_ideal(i)
        return J.change_ring(R)



r"""
Formatting utilities

This module defines helper functions that are not class methods.

AUTHORS:

- Eric Gourgoulhon, Michal Bejger (2014-2015): initial version
- Joris Vankerschaver (2010): for the function :func:`is_atomic()`
- Michael Jung (2020): extended usage of :func:`is_atomic()`
"""

# ****************************************************************************
#       Copyright (C) 2015 Eric Gourgoulhon <eric.gourgoulhon@obspm.fr>
#       Copyright (C) 2015 Michal Bejger <bejger@camk.edu.pl>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from __future__ import annotations

from typing import TYPE_CHECKING

from sage.structure.sage_object import SageObject

if TYPE_CHECKING:
    from sage.misc.latex import LatexExpr


def is_atomic(expr, sep=['+', '-']) -> bool:
    r"""
    Helper function to check whether some LaTeX expression is atomic.

    Adapted from method
    :meth:`~sage.tensor.differential_form_element.DifferentialFormFormatter._is_atomic`
    of class
    :class:`~sage.tensor.differential_form_element.DifferentialFormFormatter`
    written by Joris Vankerschaver (2010) and modified by Michael Jung (2020).

    INPUT:

    - ``expr`` -- string representing the expression (e.g. LaTeX string)
    - ``sep`` -- (default: ``['+', '-']``) a list of strings representing the
      operations (e.g. LaTeX strings)

    OUTPUT:

    - ``True`` if the operations are enclosed in parentheses and
      ``False`` otherwise.

    EXAMPLES::

        sage: from sage.tensor.modules.format_utilities import is_atomic
        sage: is_atomic("2*x")
        True
        sage: is_atomic("2+x")
        False
        sage: is_atomic("(2+x)")
        True

    Moreover the separator can be changed::

        sage: is_atomic("a*b", sep=['*'])
        False
        sage: is_atomic("(a*b)", sep=['*'])
        True
        sage: is_atomic("a mod b", sep=['mod'])
        False
        sage: is_atomic("(a mod b)", sep=['mod'])
        True

    TESTS::

        sage: is_atomic(1, sep=['*'])
        Traceback (most recent call last):
        ...
        TypeError: the argument must be a string
        sage: is_atomic("a*b", sep='*')
        Traceback (most recent call last):
        ...
        TypeError: the argument 'sep' must be a list
        sage: is_atomic("a*b", sep=[1])
        Traceback (most recent call last):
        ...
        TypeError: the argument 'sep' must consist of strings
    """
    if not isinstance(expr, str):
        raise TypeError("the argument must be a string")
    if not isinstance(sep, list):
        raise TypeError("the argument 'sep' must be a list")
    elif any(not isinstance(s, str) for s in sep):
        raise TypeError("the argument 'sep' must consist of strings")
    level = 0
    for n, c in enumerate(expr):
        if c == '(':
            level += 1
            continue
        elif c == ')':
            level -= 1
            continue
        if any(expr[n:n + len(s)] == s for s in sep):
            if level == 0 and n > 0:
                return False
    return True


def is_atomic_wedge_txt(expression) -> bool:
    r"""
    Helper function to check whether some text-formatted expression is atomic
    in terms of wedge products.

    Adapted from method
    :meth:`~sage.tensor.differential_form_element.DifferentialFormFormatter._is_atomic`
    of class
    :class:`~sage.tensor.differential_form_element.DifferentialFormFormatter`
    written by Joris Vankerschaver (2010) and modified by Michael Jung (2020).

    INPUT:

    - ``expression`` -- string representing the text-formatted expression

    OUTPUT:

    - ``True`` if wedge products are enclosed in parentheses and
      ``False`` otherwise.

    EXAMPLES::

        sage: from sage.tensor.modules.format_utilities import is_atomic_wedge_txt
        sage: is_atomic_wedge_txt("a")
        True
        sage: is_atomic_wedge_txt(r"a∧b")
        False
        sage: is_atomic_wedge_txt(r"(a∧b)")
        True
        sage: is_atomic_wedge_txt(r"(a∧b)∧c")
        False
        sage: is_atomic_wedge_txt(r"(a∧b∧c)")
        True
    """
    return is_atomic(expression, sep=['∧'])


def is_atomic_wedge_latex(expression):
    r"""
    Helper function to check whether LaTeX-formatted expression is atomic in
    terms of wedge products.

    Adapted from method
    :meth:`~sage.tensor.differential_form_element.DifferentialFormFormatter._is_atomic`
    of class
    :class:`~sage.tensor.differential_form_element.DifferentialFormFormatter`
    written by Joris Vankerschaver (2010) and modified by Michael Jung (2020).

    INPUT:

    - ``expression`` -- string representing the LaTeX expression

    OUTPUT:

    - ``True`` if wedge products are enclosed in parentheses and
      ``False`` otherwise.

    EXAMPLES::

        sage: from sage.tensor.modules.format_utilities import is_atomic_wedge_latex
        sage: is_atomic_wedge_latex(r"a")
        True
        sage: is_atomic_wedge_latex(r"a\wedge b")
        False
        sage: is_atomic_wedge_latex(r"(a\wedge b)")
        True
        sage: is_atomic_wedge_latex(r"(a\wedge b)\wedge c")
        False
        sage: is_atomic_wedge_latex(r"((a\wedge b)\wedge c)")
        True
        sage: is_atomic_wedge_latex(r"(a\wedge b\wedge c)")
        True
        sage: is_atomic_wedge_latex(r"\omega\wedge\theta")
        False
        sage: is_atomic_wedge_latex(r"(\omega\wedge\theta)")
        True
        sage: is_atomic_wedge_latex(r"\omega\wedge(\theta+a)")
        False
    """
    return is_atomic(expression, sep=['\\wedge'])


def format_mul_txt(name1, operator, name2):
    r"""
    Helper function for text-formatted names of results of multiplication or
    tensor product.

    EXAMPLES::

        sage: from sage.tensor.modules.format_utilities import format_mul_txt
        sage: format_mul_txt('a', '*', 'b')
        'a*b'
        sage: format_mul_txt('a+b', '*', 'c')
        '(a+b)*c'
        sage: format_mul_txt('a', '*', 'b+c')
        'a*(b+c)'
        sage: format_mul_txt('a+b', '*', 'c+d')
        '(a+b)*(c+d)'
        sage: format_mul_txt(None, '*', 'b')
        sage: format_mul_txt('a', '*', None)
    """
    if name1 is None or name2 is None:
        return None
    if not is_atomic(name1) or not is_atomic_wedge_txt(name1):
        name1 = '(' + name1 + ')'
    if not is_atomic(name2) or not is_atomic_wedge_txt(name2):
        name2 = '(' + name2 + ')'
    return name1 + operator + name2


def format_mul_latex(name1, operator, name2):
    r"""
    Helper function for LaTeX names of results of multiplication or tensor
    product.

    EXAMPLES::

        sage: from sage.tensor.modules.format_utilities import format_mul_latex
        sage: format_mul_latex('a', '*', 'b')
        'a*b'
        sage: format_mul_latex('a+b', '*', 'c')
        '\\left(a+b\\right)*c'
        sage: format_mul_latex('a', '*', 'b+c')
        'a*\\left(b+c\\right)'
        sage: format_mul_latex('a+b', '*', 'c+d')
        '\\left(a+b\\right)*\\left(c+d\\right)'
        sage: format_mul_latex(None, '*', 'b')
        sage: format_mul_latex('a', '*', None)
    """
    if name1 is None or name2 is None:
        return None
    if not is_atomic(name1) or not is_atomic_wedge_latex(name1):
        name1 = r'\left(' + name1 + r'\right)'
    if not is_atomic(name2) or not is_atomic_wedge_latex(name2):
        name2 = r'\left(' + name2 + r'\right)'
    return name1 + operator + name2


def format_unop_txt(operator, name):
    r"""
    Helper function for text-formatted names of results of unary operator.

    EXAMPLES::

        sage: from sage.tensor.modules.format_utilities import format_unop_txt
        sage: format_unop_txt('-', 'a')
        '-a'
        sage: format_unop_txt('-', 'a+b')
        '-(a+b)'
        sage: format_unop_txt('-', '(a+b)')
        '-(a+b)'
        sage: format_unop_txt('-', None)
    """
    if name is None:
        return None
    if not is_atomic(name) or not is_atomic_wedge_txt(name):
        # ! is_atomic_otimes_txt should be added
        name = '(' + name + ')'
    return operator + name


def format_unop_latex(operator, name):
    r"""
    Helper function for LaTeX names of results of unary operator.

    EXAMPLES::

        sage: from sage.tensor.modules.format_utilities import format_unop_latex
        sage: format_unop_latex('-', 'a')
        '-a'
        sage: format_unop_latex('-', 'a+b')
        '-\\left(a+b\\right)'
        sage: format_unop_latex('-', '(a+b)')
        '-(a+b)'
        sage: format_unop_latex('-', None)
    """
    if name is None:
        return None
    if not is_atomic(name) or not is_atomic_wedge_latex(name):
        # ! is_atomic_otimes_latex should be added
        name = r'\left(' + name + r'\right)'
    return operator + name


class FormattedExpansion(SageObject):
    r"""
    Helper class for displaying tensor expansions.

    EXAMPLES::

        sage: from sage.tensor.modules.format_utilities import FormattedExpansion
        sage: f = FormattedExpansion('v', r'\tilde v')
        sage: f
        v
        sage: latex(f)
        \tilde v
        sage: f = FormattedExpansion('x/2', r'\frac{x}{2}')
        sage: f
        x/2
        sage: latex(f)
        \frac{x}{2}
    """
    def __init__(self, txt=None, latex=None):
        r"""
        TESTS::

            sage: from sage.tensor.modules.format_utilities import FormattedExpansion
            sage: f = FormattedExpansion('v', r'\tilde v')
            sage: f
            v
        """
        self._txt = txt
        self._latex = latex

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.tensor.modules.format_utilities import FormattedExpansion
            sage: f = FormattedExpansion('v', r'\tilde v')
            sage: f._repr_()
            'v'
        """
        return self._txt

    def _latex_(self) -> LatexExpr:
        r"""
        Return a LaTeX representation of ``self``.

        EXAMPLES::

            sage: from sage.tensor.modules.format_utilities import FormattedExpansion
            sage: f = FormattedExpansion('v', r'\tilde v')
            sage: f._latex_()
            '\\tilde v'
        """
        return self._latex



# distutils: libraries = flint
# distutils: depends = flint/mpfr_mat.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void mpfr_mat_init(mpfr_mat_t mat, slong rows, slong cols, mpfr_prec_t prec) noexcept
    void mpfr_mat_clear(mpfr_mat_t mat) noexcept
    __mpfr_struct * mpfr_mat_entry(const mpfr_mat_t mat, slong i, slong j) noexcept
    void mpfr_mat_swap(mpfr_mat_t mat1, mpfr_mat_t mat2) noexcept
    void mpfr_mat_swap_entrywise(mpfr_mat_t mat1, mpfr_mat_t mat2) noexcept
    void mpfr_mat_set(mpfr_mat_t mat1, const mpfr_mat_t mat2) noexcept
    void mpfr_mat_zero(mpfr_mat_t mat) noexcept
    bint mpfr_mat_equal(const mpfr_mat_t mat1, const mpfr_mat_t mat2) noexcept
    void mpfr_mat_randtest(mpfr_mat_t mat, flint_rand_t state) noexcept
    void mpfr_mat_mul_classical(mpfr_mat_t C, const mpfr_mat_t A, const mpfr_mat_t B, mpfr_rnd_t rnd) noexcept




"""
GLPK/exact Backend (simplex method in exact rational arithmetic)

AUTHORS:

- Matthias Koeppe (2016-03)
"""

##############################################################################
#       Copyright (C) 2016 Matthias Koeppe
#  Distributed under the terms of the GNU General Public License (GPL)
#  The full text of the GPL is available at:
#                  http://www.gnu.org/licenses/
##############################################################################

cdef class GLPKExactBackend(GLPKBackend):
    """
    MIP Backend that runs the GLPK solver in exact rational simplex mode.

    The only access to data is via double-precision floats, which
    means that rationals in the input data may be rounded before
    the exact solver sees them. Thus, it is unreasonable to expect
    that arbitrary LPs with rational coefficients are solved exactly.
    Once the LP has been read into the backend, it reconstructs
    rationals from doubles and does solve exactly over the rationals,
    but results are returned as as doubles.

    There is no support for integer variables.
    """
    def __cinit__(self, maximization=True):
        """
        Constructor.

        EXAMPLES::

            sage: p = MixedIntegerLinearProgram(solver='GLPK/exact')
        """
        # inherited __cinit__ is called automatically
        import sage.numerical.backends.glpk_backend as glpk_backend
        self.solver_parameter(glpk_backend.glp_simplex_or_intopt, glpk_backend.glp_exact_simplex_only)

    cpdef int add_variable(self, lower_bound=0.0, upper_bound=None, binary=False, continuous=False, integer=False, obj=0.0, name=None) except -1:
        """
        Add a variable.

        This amounts to adding a new column to the matrix. By default,
        the variable is both nonnegative and real.

        In this backend, variables are always continuous (real).
        If integer variables are requested via the parameters
        ``binary`` and ``integer``, an error will be raised.

        INPUT:

        - ``lower_bound`` -- the lower bound of the variable (default: 0)

        - ``upper_bound`` -- the upper bound of the variable (default: ``None``)

        - ``binary`` -- ``True`` if the variable is binary (default: ``False``)

        - ``continuous`` -- ``True`` if the variable is continuous (default: ``True``)

        - ``integer`` -- ``True`` if the variable is integer (default: ``False``)

        - ``obj`` -- (optional) coefficient of this variable in the objective function (default: 0.0)

        - ``name`` -- an optional name for the newly added variable (default: ``None``)

        OUTPUT: the index of the newly created variable

        EXAMPLES::

            sage: from sage.numerical.backends.generic_backend import get_solver
            sage: p = get_solver(solver = "GLPK/exact")
            sage: p.ncols()
            0
            sage: p.add_variable()
            0
            sage: p.ncols()
            1
            sage: p.add_variable()
            1
            sage: p.add_variable(lower_bound=-2.0)
            2
            sage: p.add_variable(continuous=True)
            3
            sage: p.add_variable(name='x',obj=1.0)
            4
            sage: p.objective_coefficient(4)
            1.0

        TESTS::

            sage: p.add_variable(integer=True)
            Traceback (most recent call last):
            ...
            ValueError: GLPK/exact only supports continuous variables
            sage: p.add_variable(binary=True)
            Traceback (most recent call last):
            ...
            ValueError: GLPK/exact only supports continuous variables
        """
        if binary or integer:
            raise ValueError("GLPK/exact only supports continuous variables")
        return GLPKBackend.add_variable(self, lower_bound, upper_bound, binary, continuous,
                                        integer, obj, name)

    cpdef int add_variables(self, int number, lower_bound=0.0, upper_bound=None, binary=False, continuous=False, integer=False, obj=0.0, names=None) except -1:
        """
        Add ``number`` variables.

        This amounts to adding new columns to the matrix. By default,
        the variables are both nonnegative and real.

        In this backend, variables are always continuous (real).
        If integer variables are requested via the parameters
        ``binary`` and ``integer``, an error will be raised.

        INPUT:

        - ``n`` -- the number of new variables (must be > 0)

        - ``lower_bound`` -- the lower bound of the variable (default: 0)

        - ``upper_bound`` -- the upper bound of the variable (default: ``None``)

        - ``binary`` -- ``True`` if the variable is binary (default: ``False``)

        - ``continuous`` -- ``True`` if the variable is binary (default: ``True``)

        - ``integer`` -- ``True`` if the variable is binary (default: ``False``)

        - ``obj`` -- coefficient of all variables in the objective function (default: 0.0)

        - ``names`` -- list of names (default: ``None``)

        OUTPUT: the index of the variable created last

        EXAMPLES::

            sage: from sage.numerical.backends.generic_backend import get_solver
            sage: p = get_solver(solver = "GLPK/exact")
            sage: p.ncols()
            0
            sage: p.add_variables(5)
            4
            sage: p.ncols()
            5
            sage: p.add_variables(2, lower_bound=-2.0, obj=42.0, names=['a','b'])
            6
        """
        if binary or integer:
            raise ValueError("GLPK/exact only supports continuous variables")
        return GLPKBackend.add_variables(self, number, lower_bound, upper_bound, binary, continuous,
                                        integer, obj, names)

    cpdef set_variable_type(self, int variable, int vtype):
        """
        Set the type of a variable.

        In this backend, variables are always continuous (real).
        If integer or binary variables are requested via the parameter
        ``vtype``, an error will be raised.

        INPUT:

        - ``variable`` -- integer; the variable's id

        - ``vtype`` -- integer:

            *  1  Integer
            *  0  Binary
            * -1 Real

        EXAMPLES::

            sage: from sage.numerical.backends.generic_backend import get_solver
            sage: p = get_solver(solver = "GLPK/exact")
            sage: p.add_variables(5)
            4
            sage: p.set_variable_type(3, -1)
            sage: p.set_variable_type(3, -2)
            Traceback (most recent call last):
            ...
            ValueError: ...
        """
        if vtype != -1:
            raise ValueError("GLPK/exact only supports continuous variables")



r"""
Semistandard tableaux

This is an implementation of the abstract base class
:class:`sage.combinat.path_tableaux.path_tableau.PathTableau`.

This implementation is for semistandard tableaux, represented as a chain of partitions
(essentially, the Gelfand-Tsetlin pattern).
This generalises the jeu de taquin operations of rectification, promotion, evacuation from
standard tableaux to semistandard tableaux. The local rule is the Bender-Knuth involution.

EXAMPLES::

    sage: pt = path_tableaux.SemistandardPathTableau([[], [3], [3,2], [3,3,1],
    ....:                                             [3,3,2,1], [4,3,3,1,0]])
    sage: pt.promotion()
    [(), (2,), (3, 1), (3, 2, 1), (4, 3, 1, 0), (4, 3, 3, 1, 0)]
    sage: pt.evacuation()
    [(), (2,), (4, 0), (4, 2, 0), (4, 3, 1, 0), (4, 3, 3, 1, 0)]

    sage: pt = path_tableaux.SemistandardPathTableau([[], [3], [3,2], [3,3,1],
    ....:                                             [3,3,2,1], [9/2,3,3,1,0]])
    sage: pt.promotion()
    [(), (2,), (3, 1), (3, 2, 1), (9/2, 3, 1, 0), (9/2, 3, 3, 1, 0)]
    sage: pt.evacuation()
    [(), (5/2,), (9/2, 0), (9/2, 2, 0), (9/2, 3, 1, 0), (9/2, 3, 3, 1, 0)]

    sage: pt = path_tableaux.SemistandardPathTableau([[], [3], [4,2], [5,4,1]])
    sage: path_tableaux.CylindricalDiagram(pt)
    [       (),      (3,),    (4, 2), (5, 4, 1)]
    [         ,        (),      (3,),    (5, 2), (5, 4, 1)]
    [         ,          ,        (),      (4,),    (4, 3), (5, 4, 1)]
    [         ,          ,          ,        (),      (3,),    (5, 1), (5, 4, 1)]

    sage: pt2 = path_tableaux.SemistandardPathTableau([[3,2], [3,3,1],
    ....:                                              [3,3,2,1], [4,3,3,1,0]])
    sage: pt1 = path_tableaux.SemistandardPathTableau([[], [3], [3,2]])
    sage: pt1.commutor(pt2)
    ([(), (2,), (2, 2), (4, 2, 0)], [(4, 2, 0), (4, 3, 2, 0), (4, 3, 3, 1, 0)])
    sage: pt1.commutor(pt2,verbose=True)
    [(3, 2), (3, 3, 1), (3, 3, 2, 1), (4, 3, 3, 1, 0)]
    [(3,), (3, 2), (3, 2, 2), (4, 3, 2, 0)]
    [(), (2,), (2, 2), (4, 2, 0)]
    ([(), (2,), (2, 2), (4, 2, 0)], [(4, 2, 0), (4, 3, 2, 0), (4, 3, 3, 1, 0)])

    sage: st = SkewTableau([[None, None, None, 4, 4, 5, 6, 7], [None, 2, 4, 6, 7, 7, 7],
    ....:                   [None, 4, 5, 8, 8, 9], [None, 6, 7, 10], [None, 8, 8, 11],
    ....:                   [None], [4]])
    sage: pt = path_tableaux.SemistandardPathTableau(st)
    sage: bk = [SkewTableau(st.bender_knuth_involution(i+1)) for i in range(10)]
    sage: lr = [pt.local_rule(i+1) for i in range(10)]
    sage: [r.to_tableau() for r in lr] == bk
    True

TESTS::

    sage: pt = path_tableaux.SemistandardPathTableau([[], [3], [3,2], [3,3,1], [3,3,2,1], [4,3,3,1,0]])
    sage: TestSuite(pt).run()

    sage: pt = path_tableaux.SemistandardPathTableau([[], [3], [3,2], [3,3,1], [7/2,3,2,1], [4,3,3,1,0]])
    sage: TestSuite(pt).run()
    Failure in _test_jdt_promotion:
    Traceback (most recent call last):
    ...
    The following tests failed: _test_jdt_promotion

    sage: pt = path_tableaux.SemistandardPathTableau([[3,2], [3,3,1], [3,3,2,1], [4,3,3,1,0]])
    sage: pt.promotion()
    [(3, 2), (3, 2, 2), (4, 3, 2, 0), (4, 3, 3, 1, 0)]

AUTHORS:

- Bruce Westbury (2020): initial version
"""

#*****************************************************************************
#       Copyright (C) 2020 Bruce Westbury <bruce.westbury@gmail.com>,
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.path_tableaux.path_tableau import PathTableau, PathTableaux
from sage.combinat.combinatorial_map import combinatorial_map
from sage.combinat.skew_tableau import SkewTableau, SkewTableaux
from sage.combinat.tableau import Tableau
from sage.combinat.gelfand_tsetlin_patterns import GelfandTsetlinPattern
from sage.combinat.partition import _Partitions
from sage.rings.semirings.non_negative_integer_semiring import NN

###############################################################################


class SemistandardPathTableau(PathTableau):
    r"""
    An instance is a sequence of lists. Usually the entries will be nonnegative integers
    in which case this is the chain of partitions of a (skew) semistandard tableau.
    In general the entries are elements of an ordered abelian group; each list is weakly
    decreasing and successive lists are interleaved.

    INPUT:

    Can be any of the following

    * a sequence of partitions
    * a sequence of lists/tuples
    * a semistandard tableau
    * a semistandard skew tableau
    * a Gelfand-Tsetlin pattern

    EXAMPLES::

        sage: path_tableaux.SemistandardPathTableau([[], [2], [2,1]])
        [(), (2,), (2, 1)]

        sage: gt = GelfandTsetlinPattern([[2,1], [2]])
        sage: path_tableaux.SemistandardPathTableau(gt)
        [(), (2,), (2, 1)]

        sage: st = SemistandardTableau([[1,1], [2]])
        sage: path_tableaux.SemistandardPathTableau(st)
        [(), (2,), (2, 1)]

        sage: st = SkewTableau([[1,1], [2]])
        sage: path_tableaux.SemistandardPathTableau(st)
        [(), (2,), (2, 1)]

        sage: st = SkewTableau([[None,1,1], [2]])
        sage: path_tableaux.SemistandardPathTableau(st)
        [(1,), (3, 0), (3, 1, 0)]

        sage: path_tableaux.SemistandardPathTableau([[], [5/2], [7/2,2]])
        [(), (5/2,), (7/2, 2)]

        sage: path_tableaux.SemistandardPathTableau([[], [2.5], [3.5,2]])
        [(), (2.50000000000000,), (3.50000000000000, 2)]
    """

    @staticmethod
    def __classcall_private__(cls, st, check=True):
        r"""
        Ensure that a tableau is only ever constructed as an
        ``element_class`` call of an appropriate parent.

        EXAMPLES::

            sage: t = path_tableaux.SemistandardPathTableau([[], [2]])
            sage: t.parent()
            <sage.combinat.path_tableaux.semistandard.SemistandardPathTableaux_with_category object at ...>
        """
        return SemistandardPathTableaux()(st, check=check)

    def __init__(self, parent, st, check=True):
        r"""
        Initialize a semistandard tableau.

        TESTS::

            sage: path_tableaux.SemistandardPathTableau([(), 3, (3, 2)])
            Traceback (most recent call last):
            ...
            ValueError: [(), 3, (3, 2)] is not a sequence of lists
        """
        w = None

        if isinstance(st, SemistandardPathTableau):
            w = list(st)

        elif isinstance(st, GelfandTsetlinPattern):
            w = list(st)
            w.reverse()
            w = [(), *w]

        elif isinstance(st, (Tableau,SkewTableau)):
            w = st.to_chain()

        elif isinstance(st, (list,tuple)):
            if any(not isinstance(a,(list,tuple)) for a in st):
                raise ValueError(f"{st} is not a sequence of lists")
            w = st

        else:
            raise ValueError(f"invalid input {st} is of type {type(st)}")

        # Pad with zeroes, if necessary
        m = max(len(a)-i for i,a in enumerate(w))
        w = [list(a)+[0]*(m+i-len(a)) for i,a in enumerate(w)]
        # Convert to immutable
        w = tuple([tuple(a) for a in w])

        PathTableau.__init__(self, parent, w, check=check)

    def check(self):
        """
        Check that ``self`` is a valid path.

        EXAMPLES::

            sage: path_tableaux.SemistandardPathTableau([[], [3], [2,2]])       # indirect doctest
            Traceback (most recent call last):
            ...
            ValueError: [(), (3,), (2, 2)] does not satisfy
            the required inequalities in row 1

            sage: path_tableaux.SemistandardPathTableau([[], [3/2], [2,5/2]])   # indirect doctest
            Traceback (most recent call last):
            ...
            ValueError: [(), (3/2,), (2, 5/2)] does not satisfy
            the required inequalities in row 1


        TESTS::

            sage: path_tableaux.SemistandardPathTableau([[], [2], [1,2]])
            Traceback (most recent call last):
            ...
            ValueError: [(), (2,), (1, 2)] does not satisfy the required inequalities in row 1

            sage: path_tableaux.SemistandardPathTableau([[], [2], [1,2]], check=False)
            [(), (2,), (1, 2)]
        """
        for i in range(1,len(self)-1):
            if not all(r >= s for r,s in zip(self[i+1],self[i])):
                raise ValueError(f"{self} does not satisfy the required inequalities in row {i}")
            if not all(r >= s for r,s in zip(self[i],self[i+1][1:])):
                raise ValueError(f"{self} does not satisfy the required inequalities in row {i}")

    def size(self):
        r"""
        Return the size or length of ``self``.

        EXAMPLES::

            sage: path_tableaux.SemistandardPathTableau([[], [3], [3,2], [3,3,1], [3,3,2,1]]).size()
            5
        """
        return len(self)

    def is_skew(self):
        """
        Return ``True`` if ``self`` is skew.

        EXAMPLES::

            sage: path_tableaux.SemistandardPathTableau([[], [2]]).is_skew()
            False
            sage: path_tableaux.SemistandardPathTableau([[2,1]]).is_skew()
            True
        """
        return bool(self[0])

    def is_integral(self) -> bool:
        """
        Return ``True`` if all entries are nonnegative integers.

        EXAMPLES::

            sage: path_tableaux.SemistandardPathTableau([[], [3], [3,2]]).is_integral()
            True
            sage: path_tableaux.SemistandardPathTableau([[], [5/2], [7/2,2]]).is_integral()
            False
            sage: path_tableaux.SemistandardPathTableau([[], [3], [3,-2]]).is_integral()
            False
        """
        return all(i in NN for a in self for i in a)

    def local_rule(self, i):
        r"""
        This is the Bender-Knuth involution.

        This is implemented by toggling the entries of the `i`-th list.
        The allowed range for ``i`` is ``0 < i < len(self)-1`` so any row except
        the first and last can be changed.

        EXAMPLES::

            sage: pt = path_tableaux.SemistandardPathTableau([[], [3], [3,2],
            ....:                                             [3,3,1], [3,3,2,1]])
            sage: pt.local_rule(1)
            [(), (2,), (3, 2), (3, 3, 1), (3, 3, 2, 1)]
            sage: pt.local_rule(2)
            [(), (3,), (3, 2), (3, 3, 1), (3, 3, 2, 1)]
            sage: pt.local_rule(3)
            [(), (3,), (3, 2), (3, 2, 2), (3, 3, 2, 1)]

        TESTS::

            sage: pt = path_tableaux.SemistandardPathTableau([[], [3], [3,2],
            ....:                                             [3,3,1], [3,3,2,1]])
            sage: pt.local_rule(0)
            Traceback (most recent call last):
            ...
            ValueError: 0 is not defined on [(), (3,), (3, 2), (3, 3, 1), (3, 3, 2, 1)]
            sage: pt.local_rule(4)
            Traceback (most recent call last):
            ...
            ValueError: 4 is not defined on [(), (3,), (3, 2), (3, 3, 1), (3, 3, 2, 1)]
        """
        def toggle(i, j):
            """
            Return the toggle of entry 'self[i][j]'.
            """

            if j == 0:
                left = self[i+1][0]
            else:
                left = min(self[i+1][j], self[i-1][j-1])
            if j == len(self[i])-1:
                right = self[i+1][j+1]
            else:
                right = max(self[i+1][j+1], self[i-1][j])

            return left + right - self[i][j]

        if not 0 < i < self.size()-1:
            raise ValueError(f"{i} is not defined on {self}")

        with self.clone() as result:
            result[i] = tuple([toggle(i,k) for k in range(len(self[i]))])

        return result

    def rectify(self, inner=None, verbose=False):
        """
        Rectify ``self``.

        This gives the usual rectification of a skew standard tableau and gives a
        generalisation to skew semistandard tableaux. The usual construction uses
        jeu de taquin but here we use the Bender-Knuth involutions.

        EXAMPLES::

            sage: st = SkewTableau([[None, None, None, 4], [None, None, 1, 6],
            ....:                   [None, None, 5], [2, 3]])
            sage: path_tableaux.SemistandardPathTableau(st).rectify()
            [(), (1,), (1, 1), (2, 1, 0), (3, 1, 0, 0), (3, 2, 0, 0, 0), (4, 2, 0, 0, 0, 0)]
            sage: path_tableaux.SemistandardPathTableau(st).rectify(verbose=True)
            [[(3, 2, 2), (3, 3, 2, 0), (3, 3, 2, 1, 0), (3, 3, 2, 2, 0, 0),
              (4, 3, 2, 2, 0, 0, 0), (4, 3, 3, 2, 0, 0, 0, 0), (4, 4, 3, 2, 0, 0, 0, 0, 0)],
             [(3, 2), (3, 3, 0), (3, 3, 1, 0), (3, 3, 2, 0, 0), (4, 3, 2, 0, 0, 0),
              (4, 3, 3, 0, 0, 0, 0), (4, 4, 3, 0, 0, 0, 0, 0)],
             [(3,), (3, 1), (3, 1, 1), (3, 2, 1, 0), (4, 2, 1, 0, 0), (4, 3, 1, 0, 0, 0),
              (4, 4, 1, 0, 0, 0, 0)],
             [(), (1,), (1, 1), (2, 1, 0), (3, 1, 0, 0), (3, 2, 0, 0, 0), (4, 2, 0, 0, 0, 0)]]

        TESTS::

            sage: S = SemistandardSkewTableaux([[5,3,3], [3,1]], [3,2,2])
            sage: LHS = [path_tableaux.SemistandardPathTableau(st.rectify()) for st in S]
            sage: RHS = [path_tableaux.SemistandardPathTableau(st).rectify() for st in S]
            sage: LHS == RHS
            True

            sage: st = SkewTableau([[None, None, None, 4], [None, None, 1, 6],
            ....:                   [None, None, 5], [2, 3]])
            sage: pt = path_tableaux.SemistandardPathTableau(st)
            sage: SP = [path_tableaux.SemistandardPathTableau(it)
            ....:       for it in StandardTableaux([3,2,2])]
            sage: len(set(pt.rectify(inner=ip) for ip in SP))
            1
        """
        if not self.is_skew():
            return self

        n = len(self)
        pp = self[0]
        P = self.parent()

        if inner is None:
            initial = [pp[:r] for r in range(len(pp))]
        elif _Partitions(inner[-1]) == _Partitions(pp):
            initial = list(inner)[:-1]
        else:
            raise ValueError(f"the final shape{inner[-1]} must agree with the initial shape {pp}")

        r = len(initial)
        path = P.element_class(P, initial + list(self))
        if verbose:
            rect = [self]

        for i in range(r):
            for j in range(n-1):
                path = path.local_rule(r+j-i)
            if verbose:
                rect.append(P.element_class(P, list(path)[r-i-1:r+n-i-1]))

        if verbose:
            return rect
        return P.element_class(P, list(path)[:n])

    @combinatorial_map(name='to semistandard tableau')
    def to_tableau(self):
        r"""
        Convert ``self`` to a :class:`SemistandardTableau`.

        The :class:`SemistandardSkewTableau` is not implemented so this returns a :class:`SkewTableau`

        EXAMPLES::

            sage: pt = path_tableaux.SemistandardPathTableau([[], [3], [3,2], [3,3,1],
            ....:                                             [3,3,2,1], [4,3,3,1,0]])
            sage: pt.to_tableau()
            [[1, 1, 1, 5], [2, 2, 3], [3, 4, 5], [4]]

        TESTS::

            sage: SST = SemistandardTableaux(shape=[5,5,3], eval=[2,2,3,4,2])
            sage: all(st == path_tableaux.SemistandardPathTableau(st).to_tableau()      # needs sage.modules
            ....:     for st in SST)
            True
        """
        from sage.combinat.tableau import from_chain

        if not self.is_integral():
            raise ValueError(f"{self} must have all entries nonnegative integers")

        lt = [[i for i in a if i > 0] for a in self]
        if self.is_skew():
            return SkewTableaux().from_chain(lt)
        return from_chain(lt)

    @combinatorial_map(name='to Gelfand-Tsetlin pattern')
    def to_pattern(self):
        r"""
        Convert ``self`` to a Gelfand-Tsetlin pattern.

        EXAMPLES::

            sage: pt = path_tableaux.SemistandardPathTableau([[], [3], [3,2], [3,3,1],
            ....:                                             [3,3,2,1], [4,3,3,1]])
            sage: pt.to_pattern()
            [[4, 3, 3, 1, 0], [3, 3, 2, 1], [3, 3, 1], [3, 2], [3]]

        TESTS::

            sage: pt = path_tableaux.SemistandardPathTableau([[3,2], [3,3,1], [3,3,2,1], [4,3,3,1]])
            sage: pt.to_pattern()
            Traceback (most recent call last):
            ...
            ValueError: [(3, 2), (3, 3, 1), (3, 3, 2, 1), (4, 3, 3, 1, 0)] cannot be a skew tableau

            sage: GT = GelfandTsetlinPatterns(top_row=[5,5,3])
            sage: all(gt == path_tableaux.SemistandardPathTableau(gt).to_pattern() for gt in GT)
            True

            sage: GT = GelfandTsetlinPatterns(top_row=[5,5,3])
            sage: all(gt.to_tableau() == path_tableaux.SemistandardPathTableau(gt).to_tableau() for gt in GT)
            True
        """
        if self.is_skew():
            raise ValueError(f"{self} cannot be a skew tableau")

        lt = list(self)
        lt.reverse()
        if not lt[-1]:
            lt.pop()

        return GelfandTsetlinPattern([list(a) for a in lt])

    def _test_jdt_promotion(self, **options):
        """
        Check that promotion agrees with :meth:`Tableau.promotion_inverse`
        constructed using jeu de taquin.

        TESTS::

            sage: pt = path_tableaux.SemistandardPathTableau([(), (1,), (2,1), (4,2),
            ....:                                             (4,3,1), (4,3,3)])
            sage: pt._test_jdt_promotion()

            sage: pt = path_tableaux.SemistandardPathTableau([(), (1,), (2,1), (4,2),
            ....:                                             (4,3,1), (9/2,3,3)])
            sage: pt._test_jdt_promotion()
            Traceback (most recent call last):
            ...
            ValueError: [(), (1,), (2, 1), (4, 2, 0), (4, 3, 1, 0), (9/2, 3, 3, 0, 0)]
            must have all entries nonnegative integers
        """
        if not self.is_integral():
            raise ValueError(f"{self} must have all entries nonnegative integers")

        tester = self._tester(**options)
        LHS = self.promotion().to_tableau()
        RHS = self.to_tableau().promotion_inverse(len(self)-2)
        tester.assertEqual(LHS,RHS)


class SemistandardPathTableaux(PathTableaux):
    """
    The parent class for :class:`SemistandardPathTableau`.
    """

    def _an_element_(self):
        """
        Return an element of ``self``.

        EXAMPLES::

            sage: path_tableaux.SemistandardPathTableaux()._an_element_()
            [(), (2,), (2, 1)]
        """
        return SemistandardPathTableau([[], [2], [2,1]])

    Element = SemistandardPathTableau



from sage.libs.gmp.mpz cimport mpz_t
from sage.libs.flint.types cimport padic_ctx_t, fmpz_t, fmpz_poly_t
from sage.rings.padics.pow_computer cimport PowComputer_class

cdef class PowComputer_flint(PowComputer_class):
    cdef padic_ctx_t ctx
    cdef fmpz_t fprime
    cdef fmpz_t half_prime
    cdef fmpz_t _fpow_array
    cdef fmpz_t _fpow_variable
    cdef mpz_t top_power

    cdef fmpz_t* pow_fmpz_t_tmp(self, unsigned long n) except NULL
    cdef unsigned long capdiv(self, unsigned long n) noexcept

    cdef fmpz_t tfmpz

cdef class PowComputer_flint_1step(PowComputer_flint):
    cdef fmpz_t q
    cdef fmpz_poly_t modulus
    cdef fmpz_poly_t powhelper_oneunit
    cdef fmpz_poly_t powhelper_teichdiff
    cdef fmpz_poly_t* _moduli
    cdef fmpz_poly_t* get_modulus(self, unsigned long n) noexcept
    cdef fmpz_poly_t* get_modulus_capdiv(self, unsigned long n) noexcept
    cdef _new_fmpz_poly(self, fmpz_poly_t value, var=*)

cdef class PowComputer_flint_unram(PowComputer_flint_1step):
    # WARNING:
    # These variables are modified by the linkage and must no be used anywhere else
    # (other than in __(c)init__)
    cdef fmpz_t fmpz_ccmp, fmpz_cval, fmpz_cinv, fmpz_cinv2, fmpz_cexp, fmpz_ctm, fmpz_cconv
    cdef fmpz_poly_t poly_cconv, poly_ctm, poly_ccmp, poly_cinv, poly_cisunit, poly_cinv2, poly_flint_rep, poly_matmod, shift_rem, aliasing
    cdef mpz_t mpz_cpow, mpz_ctm, mpz_cconv, mpz_matmod

cdef class PowComputer_flint_eis(PowComputer_flint_1step):
    pass



# sage.doctest: optional - pycryptosat, needs sage.modules brial
"""
SAT Functions for Boolean Polynomials

These highlevel functions support solving and learning from Boolean polynomial systems. In this
context, "learning" means the construction of new polynomials in the ideal spanned by the original
polynomials.

AUTHOR:

- Martin Albrecht (2012): initial version

Functions
^^^^^^^^^
"""
# ###########################################################################
#  Copyright (C) 2012 Martin Albrecht <martinralbrecht@googlemail.com>
#  Distributed under the terms of the GNU General Public License (GPL)
#  The full text of the GPL is available at:
#                  https://www.gnu.org/licenses/
# ###########################################################################
from sage.rings.polynomial.multi_polynomial_sequence import PolynomialSequence
from sage.sat.converters import ANF2CNFConverter
from sage.sat.solvers import SatSolver


def solve(F, converter=None, solver=None, n=1, target_variables=None, **kwds):
    """
    Solve system of Boolean polynomials ``F`` by solving the
    SAT-problem -- produced by ``converter`` -- using ``solver``.

    INPUT:

    - ``F`` -- a sequence of Boolean polynomials

    - ``n`` -- number of solutions to return. If ``n`` is +infinity
      then all solutions are returned. If ``n <infinity`` then ``n``
      solutions are returned if ``F`` has at least ``n``
      solutions. Otherwise, all solutions of ``F`` are
      returned. (default: ``1``)

    - ``converter`` -- an ANF to CNF converter class or object.  If
      ``converter`` is ``None`` then
      :class:`sage.sat.converters.polybori.CNFEncoder` is used to
      construct a new converter. (default: ``None``)

    - ``solver`` -- a SAT-solver class or object. If ``solver`` is
      ``None`` then :class:`sage.sat.solvers.cryptominisat.CryptoMiniSat`
      is used to construct a new converter.  (default: ``None``)

    - ``target_variables`` -- list of variables. The elements of the list are
      used to exclude a particular combination of variable assignments of a
      solution from any further solution. Furthermore ``target_variables``
      denotes which variable-value pairs appear in the solutions. If
      ``target_variables`` is ``None`` all variables appearing in the
      polynomials of ``F`` are used to construct exclusion clauses.
      (default: ``None``)

    - ``**kwds`` -- parameters can be passed to the converter and the
      solver by prefixing them with ``c_`` and ``s_`` respectively. For
      example, to increase CryptoMiniSat's verbosity level, pass
      ``s_verbosity=1``.

    OUTPUT:

    A list of dictionaries, each of which contains a variable assignment
    solving ``F``.

    EXAMPLES:

    We construct a very small-scale AES system of equations::

        sage: sr = mq.SR(1, 1, 1, 4, gf2=True, polybori=True)
        sage: while True:  # workaround (see :issue:`31891`)
        ....:     try:
        ....:         F, s = sr.polynomial_system()
        ....:         break
        ....:     except ZeroDivisionError:
        ....:         pass

    and pass it to a SAT solver::

        sage: from sage.sat.boolean_polynomials import solve as solve_sat
        sage: s = solve_sat(F)
        sage: F.subs(s[0])
        Polynomial Sequence with 36 Polynomials in 0 Variables

    This time we pass a few options through to the converter and the solver::

        sage: s = solve_sat(F, c_max_vars_sparse=4, c_cutting_number=8)
        sage: F.subs(s[0])
        Polynomial Sequence with 36 Polynomials in 0 Variables

    We construct a very simple system with three solutions
    and ask for a specific number of solutions::

        sage: B.<a,b> = BooleanPolynomialRing()
        sage: f = a*b
        sage: l = solve_sat([f],n=1)
        sage: len(l) == 1, f.subs(l[0])
        (True, 0)

        sage: l = solve_sat([a*b],n=2)
        sage: len(l) == 2, f.subs(l[0]), f.subs(l[1])
        (True, 0, 0)

        sage: sorted((d[a], d[b]) for d in solve_sat([a*b], n=3))
        [(0, 0), (0, 1), (1, 0)]
        sage: sorted((d[a], d[b]) for d in solve_sat([a*b], n=4))
        [(0, 0), (0, 1), (1, 0)]
        sage: sorted((d[a], d[b]) for d in solve_sat([a*b], n=infinity))
        [(0, 0), (0, 1), (1, 0)]

    In the next example we see how the ``target_variables`` parameter works::

        sage: from sage.sat.boolean_polynomials import solve as solve_sat
        sage: R.<a,b,c,d> = BooleanPolynomialRing()
        sage: F = [a + b, a + c + d]

    First the normal use case::

        sage: sorted((D[a], D[b], D[c], D[d])
        ....:        for D in solve_sat(F, n=infinity))
        [(0, 0, 0, 0), (0, 0, 1, 1), (1, 1, 0, 1), (1, 1, 1, 0)]

    Now we are only interested in the solutions of the variables a and b::

        sage: solve_sat(F, n=infinity, target_variables=[a,b])
        [{b: 0, a: 0}, {b: 1, a: 1}]

    Here, we generate and solve the cubic equations of the AES SBox (see :issue:`26676`)::

        sage: # long time
        sage: from sage.rings.polynomial.multi_polynomial_sequence import PolynomialSequence
        sage: from sage.sat.boolean_polynomials import solve as solve_sat
        sage: sr = sage.crypto.mq.SR(1, 4, 4, 8,
        ....:                        allow_zero_inversions=True)
        sage: sb = sr.sbox()
        sage: eqs = sb.polynomials(degree=3)
        sage: eqs = PolynomialSequence(eqs)
        sage: variables = map(str, eqs.variables())
        sage: variables = ",".join(variables)
        sage: R = BooleanPolynomialRing(16, variables)
        sage: eqs = [R(eq) for eq in eqs]
        sage: sls_aes = solve_sat(eqs, n=infinity)
        sage: len(sls_aes)
        256

    TESTS:

    Test that :issue:`26676` is fixed::

        sage: varl = ['k{0}'.format(p) for p in range(29)]
        sage: B = BooleanPolynomialRing(names=varl)
        sage: B.inject_variables(verbose=False)
        sage: keqs = [
        ....:     k0 + k6 + 1,
        ....:     k3 + k9 + 1,
        ....:     k5*k18 + k6*k18 + k7*k16 + k7*k10,
        ....:     k9*k17 + k8*k24 + k11*k17,
        ....:     k1*k13 + k1*k15 + k2*k12 + k3*k15 + k4*k14,
        ....:     k5*k18 + k6*k16 + k7*k18,
        ....:     k3 + k26,
        ....:     k0 + k19,
        ....:     k9 + k28,
        ....:     k11 + k20]
        sage: from sage.sat.boolean_polynomials import solve as solve_sat
        sage: solve_sat(keqs, n=1, solver=SAT('cryptominisat'))
        [{k28: 0,
          k26: 1,
          k24: 0,
          k20: 0,
          k19: 0,
          k18: 0,
          k17: 0,
          k16: 0,
          k15: 0,
          k14: 0,
          k13: 0,
          k12: 0,
          k11: 0,
          k10: 0,
          k9: 0,
          k8: 0,
          k7: 0,
          k6: 1,
          k5: 0,
          k4: 0,
          k3: 1,
          k2: 0,
          k1: 0,
          k0: 0}]
        sage: solve_sat(keqs, n=1, solver=SAT('picosat'))                   # optional - pycosat
        [{k28: 0,
          k26: 1,
          k24: 0,
          k20: 0,
          k19: 0,
          k18: 0,
          k17: 0,
          k16: 0,
          k15: 0,
          k14: 0,
          k13: 1,
          k12: 1,
          k11: 0,
          k10: 0,
          k9: 0,
          k8: 0,
          k7: 0,
          k6: 1,
          k5: 0,
          k4: 1,
          k3: 1,
          k2: 1,
          k1: 1,
          k0: 0}]

    .. NOTE::

        Although supported, passing converter and solver objects
        instead of classes is discouraged because these objects are
        stateful.
    """
    assert n > 0

    try:
        len(F)
    except AttributeError:
        F = F.gens()
        len(F)

    P = next(iter(F)).parent()
    K = P.base_ring()

    if target_variables is None:
        target_variables = PolynomialSequence(F).variables()
    else:
        target_variables = PolynomialSequence(target_variables).variables()
        assert set(target_variables).issubset(set(P.gens()))

    # instantiate the SAT solver

    if solver is None:
        from sage.sat.solvers import CryptoMiniSat as solver

    if not isinstance(solver, SatSolver):
        solver_kwds = {k[2:]: v for k, v in kwds.items()
                       if k.startswith("s_")}

        solver = solver(**solver_kwds)

    # instantiate the ANF to CNF converter

    if converter is None:
        from sage.sat.converters.polybori import CNFEncoder as converter

    if not isinstance(converter, ANF2CNFConverter):
        converter_kwds = {k[2:]: v for k, v in kwds.items()
                          if k.startswith("c_")}

        converter = converter(solver, P, **converter_kwds)

    phi = converter(F)
    rho = {phi[i]: i for i in range(len(phi))}

    S = []

    while True:
        s = solver()

        if s:
            S.append({x: K(s[rho[x]]) for x in target_variables})

            if n is not None and len(S) == n:
                break

            exclude_solution = tuple(-rho[x] if s[rho[x]] else rho[x] for x in target_variables)
            solver.add_clause(exclude_solution)

        else:
            try:
                learnt = solver.learnt_clauses(unitary_only=True)
                if learnt:
                    S.append({phi[abs(i) - 1]: K(i < 0) for i in learnt})
                else:
                    S.append(s)
                    break
            except (AttributeError, NotImplementedError):
                # solver does not support recovering learnt clauses
                S.append(s)
                break

    if len(S) == 1:
        if S[0] is False:
            return False
        if S[0] is None:
            return None
    elif S[-1] is False:
        return S[0:-1]
    return S


def learn(F, converter=None, solver=None, max_learnt_length=3, interreduction=False, **kwds):
    """
    Learn new polynomials by running SAT-solver ``solver`` on
    SAT-instance produced by ``converter`` from ``F``.

    INPUT:

    - ``F`` -- a sequence of Boolean polynomials

    - ``converter`` -- an ANF to CNF converter class or object.  If ``converter`` is ``None`` then
      :class:`sage.sat.converters.polybori.CNFEncoder` is used to construct a new
      converter. (default: ``None``)

    - ``solver`` -- a SAT-solver class or object. If ``solver`` is ``None`` then
      :class:`sage.sat.solvers.cryptominisat.CryptoMiniSat` is used to construct a new converter.
      (default: ``None``)

    - ``max_learnt_length`` -- only clauses of length <= ``max_length_learnt`` are considered and
      converted to polynomials. (default: ``3``)

    - ``interreduction`` -- inter-reduce the resulting polynomials (default: ``False``)

    .. NOTE::

        More parameters can be passed to the converter and the solver by prefixing them with ``c_`` and
        ``s_`` respectively. For example, to increase CryptoMiniSat's verbosity level, pass
        ``s_verbosity=1``.

    OUTPUT: a sequence of Boolean polynomials

    EXAMPLES::

        sage: from sage.sat.boolean_polynomials import learn as learn_sat

    We construct a simple system and solve it::

        sage: set_random_seed(2300)
        sage: sr = mq.SR(1, 2, 2, 4, gf2=True, polybori=True)
        sage: F,s = sr.polynomial_system()
        sage: H = learn_sat(F)
        sage: H[-1]
        k033 + 1
    """
    try:
        len(F)
    except AttributeError:
        F = F.gens()
        len(F)

    P = next(iter(F)).parent()
    K = P.base_ring()

    # instantiate the SAT solver

    if solver is None:
        from sage.sat.solvers.cryptominisat import CryptoMiniSat as solver

    solver_kwds = {k[2:]: v for k, v in kwds.items()
                   if k.startswith("s_")}

    solver = solver(**solver_kwds)

    # instantiate the ANF to CNF converter

    if converter is None:
        from sage.sat.converters.polybori import CNFEncoder as converter

    converter_kwds = {k[2:]: v for k, v in kwds.items()
                      if k.startswith("c_")}

    converter = converter(solver, P, **converter_kwds)

    phi = converter(F)
    rho = {phi[i]: i for i in range(len(phi))}

    s = solver()

    if s:
        learnt = [x + K(s[rho[x]]) for x in P.gens()]
    else:
        learnt = []
        try:
            lc = solver.learnt_clauses()
        except (AttributeError, NotImplementedError):
            # solver does not support recovering learnt clauses
            lc = []
        for c in lc:
            if len(c) <= max_learnt_length:
                try:
                    learnt.append(converter.to_polynomial(c))
                except (ValueError, NotImplementedError, AttributeError):
                    # the solver might have learnt clauses that contain CNF
                    # variables which have no correspondence to variables in our
                    # polynomial ring (XOR chaining variables for example)
                    pass

    learnt = PolynomialSequence(P, learnt)

    if interreduction:
        learnt = learnt.ideal().interreduced_basis()
    return learnt



r"""
Rational point sets of Jacobians of genus-2 curves (split case)

AUTHORS:

- Sabrina Kunzweiler, Gareth Ma, Giacomo Pope (2024): adapt to smooth model
"""

# ****************************************************************************
#       Copyright (C) 2025 Sabrina Kunzweiler <sabrina.kunzweiler@math.u-bordeaux.fr>
#                     2025 Gareth Ma <grhkm21@gmail.com>
#                     2025 Giacomo Pope <giacomopope@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.schemes.hyperelliptic_curves.jacobian_homset_split import (
    HyperellipticJacobianHomsetSplit,
)


class HyperellipticJacobianHomsetSplit_g2(HyperellipticJacobianHomsetSplit):
    r"""
    Special class to handle optimisations for jacobian homset computations
    in genus two for hyperlliptic curves with a split model
    """

    pass
